# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
import numpy as np
import time
from bark.runtime.commons import ParameterServer
from bark.runtime.viewer import MPViewer
from bark.runtime.commons import XodrParser
from bark.core.models.execution import ExecutionModelInterpolate
from bark.core.models.dynamic import SingleTrackModel, StateDefinition
from bark.core.world import World, MakeTestWorldHighway
from bark.core.world.goal_definition import GoalDefinitionPolygon, GoalDefinitionStateLimitsFrenet
from bark.core.world.agent import Agent
from bark.core.world.map import MapInterface, Roadgraph
from bark.core.geometry.standard_shapes import CarLimousine, CarRectangle
from bark.core.geometry import Point2d, Polygon2d, Line2d
from bark.core.world.evaluation import EvaluatorDrivableArea
from bark.core.world.opendrive import OpenDriveMap, XodrRoad, PlanView, \
    MakeXodrMapOneRoadTwoLanes, XodrLaneSection, XodrLane
from bark.runtime.viewer.video_renderer import VideoRenderer
from bark.core.models.behavior import * 
import os


class PythonDistanceBehavior(BehaviorModel):
  """Python behavior model to give an example of observed world interfaces
  """
  def __init__(self, params = None):
    BehaviorModel.__init__(
      self, params)
    self._params = params
    self._distance_range_constant_velocity = \
            self._params["PythonDistanceBehavior::RangeConstantVelocity", \
           "Range in meters defining when controlled vehicle does not change velocity", [10, 20]]

  def Plan(self, delta_time, observed_world):
    # Get state of observer
    ego_agent = observed_world.ego_agent
    ego_agent_state = observed_world.ego_state # or via ego_agent.state
    ego_velocity = ego_agent_state[int(StateDefinition.VEL_POSITION)]
    ego_x = ego_agent_state[int(StateDefinition.X_POSITION)]
    ego_y = ego_agent_state[int(StateDefinition.Y_POSITION)]
    ego_theta = ego_agent_state[int(StateDefinition.THETA_POSITION)]
    print("Ego x: {}, y: {}, v: {}, theta: {}".format(
        ego_x, ego_y, ego_velocity, ego_theta))

    # get distance and state of observer's leading vehicle
    # 1) get tuple with agent and frenet position
    leading_agent_pair = observed_world.GetAgentInFront() # to get agent behind use GetAgentBehind()
    
    # 2) get state of leading agent from tuple
    leading_agent = leading_agent_pair[0]
    leading_agent_state = leading_agent.state
    leading_agent_x = leading_agent_state[int(StateDefinition.X_POSITION)]
    # to get other state values, same procedure as for ego agent
    # ....
    # 3) get frenet longitudinal distance to leading agent from tuple
    leading_frenet = leading_agent_pair[1]
    longitudinal_dist = leading_frenet.lon
    # 4) subtract vehicle shapes rear front to get true physical distance
    vehicle_margins = ego_agent.shape.front_dist + leading_agent.shape.rear_dist
    longitudinal_dist_margins = longitudinal_dist - vehicle_margins

    print("Leading x: {}, ... long dist: {}, long dist margins: {}".format(
        leading_agent_x, longitudinal_dist, longitudinal_dist_margins))

    # select some action 
    # we choose a very simple heuristic to select among actions 
    # decelerate, constant velocity or accelerate
    acceleration = None
    if longitudinal_dist_margins < self._distance_range_constant_velocity[0]:
      acceleration = -4.0
    elif longitudinal_dist_margins > self._distance_range_constant_velocity[0] and \
          longitudinal_dist_margins < self._distance_range_constant_velocity[1]:
      acceleration = 0.0
    else:
      acceleration = 4.0

    # a motion primitive model converts it to trajectory
    behavior = PrimitiveConstAccStayLane(self._params, acceleration)
    traj = behavior.Plan(delta_time, observed_world, observed_world.lane_corridor)

    # set internal behavior parameters
    super(PythonDistanceBehavior, self).SetLastTrajectory(traj)
    super(PythonDistanceBehavior, self).SetLastAction(acceleration)
    print("Trajectory: {}".format(traj))
    return traj

  def Clone(self):
    return self



class SystemTests(unittest.TestCase):
    """ This shall serve as a full system test, importing world, agent, and behavior models
    """
    #@unittest.skip
    def test_uct_single_agent(self):
        try:
            from bark.core.models.behavior import BehaviorUCTSingleAgentMacroActions
        except:
            print("Rerun with --define planner_uct=true")
            return
        # World Definition
        scenario_param_file ="macro_actions_test.json" # must be within examples params folder
        params = ParameterServer(filename= os.path.join(os.path.dirname(__file__),"params/",scenario_param_file))

        world = World(params)

        # Model Definitions
        behavior_model = BehaviorUCTSingleAgentMacroActions(params)
        execution_model = ExecutionModelInterpolate(params)
        dynamic_model = SingleTrackModel(params)

        behavior_model2 = BehaviorConstantVelocity(params)
        execution_model2 = ExecutionModelInterpolate(params)
        dynamic_model2 = SingleTrackModel(params)

        # Map Definition
        map_interface = MapInterface()
        xodr_map = MakeXodrMapOneRoadTwoLanes()
        map_interface.SetOpenDriveMap(xodr_map)
        world.SetMap(map_interface)

        # agent_2d_shape = CarLimousine()
        agent_2d_shape = CarRectangle()
        init_state = np.array([0, 3, -5.25, 0, 20])
        agent_params = params.AddChild("agent1")


        # goal_polygon = Polygon2d(
        #     [1, 1, 0], [Point2d(0, 0), Point2d(0, 2), Point2d(2, 2), Point2d(2, 0)])
        # goal_definition = GoalDefinitionPolygon(goal_polygon)
        # goal_polygon = goal_polygon.Translate(Point2d(90, -2))

        center_line = Line2d()
        center_line.AddPoint(Point2d(0.0, -1.75))
        center_line.AddPoint(Point2d(100.0, -1.75))

        max_lateral_dist = (0.4,0.5)
        max_orientation_diff = (0.08, 0.1)
        velocity_range = (5.0, 20.0)
        goal_definition = GoalDefinitionStateLimitsFrenet(center_line,
                        max_lateral_dist, max_orientation_diff,
                        velocity_range)

        agent = Agent(init_state, behavior_model, dynamic_model, execution_model,
                      agent_2d_shape, agent_params, goal_definition, map_interface)
        world.AddAgent(agent)

        init_state2 = np.array([0, 25, -5.25, 0, 0])
        agent2 = Agent(init_state2, behavior_model2, dynamic_model2, execution_model2,
                        agent_2d_shape, agent_params, goal_definition, map_interface)
        world.AddAgent(agent2)

        # viewer
        viewer = MPViewer(params=params, use_world_bounds=True)

        # World Simulation
        sim_step_time = params["simulation"]["step_time",
                                              "Step-time in simulation", 0.2]
        sim_real_time_factor = params["simulation"]["real_time_factor",
                                                    "execution in real-time or faster", 1]

        # Draw map
        video_renderer = VideoRenderer(renderer=viewer, world_step_time=sim_step_time)

        for _ in range(0, 5):
            world.Step(sim_step_time)
            viewer.clear()
            video_renderer.drawWorld(world)
            video_renderer.drawGoalDefinition(goal_definition)
            time.sleep(sim_step_time/sim_real_time_factor)

        video_renderer.export_video(filename="./test_video_intermediate", remove_image_dir=True)


    def test_python_behavior_model(self):
        # World Definition
        scenario_param_file ="macro_actions_test.json" # must be within examples params folder
        params = ParameterServer(filename= os.path.join(os.path.dirname(__file__),"params/",scenario_param_file))

        world = World(params)

        # Define two behavior models one python one standard c++ model
        behavior_model = PythonDistanceBehavior(params)
        execution_model = ExecutionModelInterpolate(params)
        dynamic_model = SingleTrackModel(params)

        behavior_model2 = BehaviorConstantVelocity(params)
        execution_model2 = ExecutionModelInterpolate(params)
        dynamic_model2 = SingleTrackModel(params)
        
        # Define the map interface and load a testing map
        map_interface = MapInterface()
        xodr_map = MakeXodrMapOneRoadTwoLanes()
        map_interface.SetOpenDriveMap(xodr_map)
        world.SetMap(map_interface)

        # Define the agent shapes
        agent_2d_shape = CarRectangle()
        init_state = np.array([0, 3, -5.25, 0, 20])
        
        # Define the goal definition for agents
        center_line = Line2d()
        center_line.AddPoint(Point2d(0.0, -1.75))
        center_line.AddPoint(Point2d(100.0, -1.75))

        max_lateral_dist = (0.4,0.5)
        max_orientation_diff = (0.08, 0.1)
        velocity_range = (5.0, 20.0)
        goal_definition = GoalDefinitionStateLimitsFrenet(center_line,
                        max_lateral_dist, max_orientation_diff,
                        velocity_range)

        # define two agents with the different behavior models
        agent_params = params.AddChild("agent1")
        agent = Agent(init_state, behavior_model, dynamic_model, execution_model,
                      agent_2d_shape, agent_params, goal_definition, map_interface)
        world.AddAgent(agent)

        init_state2 = np.array([0, 25, -5.25, 0, 15])
        agent2 = Agent(init_state2, behavior_model2, dynamic_model2, execution_model2,
                        agent_2d_shape, agent_params, goal_definition, map_interface)
        world.AddAgent(agent2)

        # viewer
        viewer = MPViewer(params=params, use_world_bounds=True)

        # World Simulation
        sim_step_time = params["simulation"]["step_time",
                                              "Step-time in simulation", 0.2]
        sim_real_time_factor = params["simulation"]["real_time_factor",
                                                    "execution in real-time or faster", 1]

        # Draw map
        video_renderer = VideoRenderer(renderer=viewer, world_step_time=sim_step_time)

        for _ in range(0, 20):
            world.Step(sim_step_time)
            viewer.clear()
            video_renderer.drawWorld(world)
            video_renderer.drawGoalDefinition(goal_definition, "red", 0.5, "red")
            time.sleep(sim_step_time/sim_real_time_factor)

        video_renderer.export_video(filename="./test_video_intermediate", remove_image_dir=True)

if __name__ == '__main__':
    unittest.main()



import sys,os

sys.path.append(os.path.join(os.path.dirname(__file__),'../external'))
sys.path.append(os.path.join(os.path.dirname(__file__),'../external/benchmark_database'))
sys.path.append(os.path.join(os.path.dirname(__file__),'../external/com_github_interaction_dataset_interaction_dataset/python'))




# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.scenario.interaction_dataset_processing.dataset_decomposer import DatasetDecomposer
import os

class DatasetDecomposerTest(unittest.TestCase):
    def test_decompose_dataset(self):

        map_filename = os.path.join(os.path.dirname(__file__),"data/DR_DEU_Merging_MT_v01_shifted.xodr")
        track_filename = os.path.join(os.path.dirname(__file__),"data/interaction_dataset_dummy_track.csv")

        dataset_decomposer = DatasetDecomposer(
            map_filename=map_filename, track_filename=track_filename)

        dataset_decomposer.decompose()



if __name__ == '__main__':
    unittest.main()




# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.core.world.agent import Agent
from bark.core.models.behavior import BehaviorStaticTrajectory, BehaviorMobil
from bark.core.models.dynamic import StateDefinition
from bark.core.world.goal_definition import GoalDefinition, GoalDefinitionPolygon
from bark.core.geometry import Point2d, Polygon2d, Norm0To2PI
from bark.runtime.commons.model_json_conversion import ModelJsonConversion
# Interaction dataset tools
from com_github_interaction_dataset_interaction_dataset.python.utils import dataset_reader
from com_github_interaction_dataset_interaction_dataset.python.utils import dict_utils
import numpy as np

from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo


def bark_state_from_motion_state(state, time_offset=0):
    bark_state = np.zeros(int(StateDefinition.MIN_STATE_SIZE))
    bark_state[int(StateDefinition.TIME_POSITION)] = (
        state.time_stamp_ms - time_offset) / 1000.0
    bark_state[int(StateDefinition.X_POSITION)] = state.x
    bark_state[int(StateDefinition.Y_POSITION)] = state.y
    bark_state[int(StateDefinition.THETA_POSITION)] = Norm0To2PI(state.psi_rad)
    bark_state[int(StateDefinition.VEL_POSITION)] = pow(
        pow(state.vx, 2) + pow(state.vy, 2), 0.5)
    return bark_state.reshape((1, int(StateDefinition.MIN_STATE_SIZE)))


def trajectory_from_track(track, start=0, end=None):
    states = list(dict_utils.get_item_iterator(track.motion_states))
    if end is None:
        end = states[-1][0]
    filtered_motion_states = list(
        filter(lambda s: start <= s[0] <= end, states))
    n = len(filtered_motion_states)
    traj = np.zeros((n, int(StateDefinition.MIN_STATE_SIZE)))
    for i, state in enumerate(filtered_motion_states):
        # TODO: rename start to be the scenario start time!
        traj[i, :] = bark_state_from_motion_state(state[1], start)
    return traj


def shape_from_track(track, wheelbase=2.7):
    offset = wheelbase / 2.0
    length = track.length
    width = track.width
    pose = [0.0, 0.0, 0.0]
    points = [[length / 2.0 + offset, -width / 2.0], [length / 2.0 + offset, width / 2.0], [-length / 2.0 + offset, width / 2.0],
              [-length / 2.0 + offset, -width / 2.0], [length / 2.0 + offset, -width / 2.0]]
    poly = Polygon2d(pose, points)
    return poly


def init_state_from_track(track, start):
    minimum_start = min(track.motion_states)
    if minimum_start > start:
        start = minimum_start
    state = track.motion_states[start]
    return bark_state_from_motion_state(state, state.time_stamp_ms).reshape((int(StateDefinition.MIN_STATE_SIZE), 1))


def goal_definition_from_track(track, end):
    states = list(dict_utils.get_item_iterator(track.motion_states))
    motion_state = states[-1][1]
    bark_state = bark_state_from_motion_state(motion_state)
    goal_polygon = Polygon2d(np.array([0.0, 0.0, 0.0]),
                             [Point2d(-1.5, 0),
                              Point2d(-1.5, 8),
                              Point2d(1.5, 8),
                              Point2d(1.5, 0)])
    goal_polygon = goal_polygon.Translate(Point2d(bark_state[0, int(StateDefinition.X_POSITION)],
                                                  bark_state[0, int(StateDefinition.Y_POSITION)]))
    goal_definition = GoalDefinitionPolygon(goal_polygon)
    return goal_definition


def behavior_from_track(track, params, start, end):
    return BehaviorStaticTrajectory(params, trajectory_from_track(track, start, end))


def track_from_trackfile(filename, track_id):
    track_dictionary = dataset_reader.read_tracks(filename)
    track = track_dictionary[track_id]
    # TODO: Filter track
    return track


def agent_from_trackfile(track_params, param_server, scenario_track_info, agent_id):
    if scenario_track_info.GetEgoTrackInfo().GetTrackId() == agent_id:
      agent_track_info = scenario_track_info.GetEgoTrackInfo()
    elif agent_id in scenario_track_info.GetOtherTrackInfos().keys():
      agent_track_info = scenario_track_info.GetOtherTrackInfos()[agent_id]
    else:
      raise ValueError("unknown agent id {}".format(agent_id))
    fname = agent_track_info.GetFileName()  # track_params["filename"]
    track_id = agent_track_info.GetTrackId()  # track_params["track_id"]
    agent_id = agent_track_info.GetTrackId()
    track = track_from_trackfile(fname, track_id)
    start_time = scenario_track_info.GetStartTs()
    end_time = scenario_track_info.GetEndTs()

    behavior_model = track_params["behavior_model"]
    model_converter = ModelJsonConversion()
    if behavior_model is None:
        # each agent need's its own param server
        behavior = behavior_from_track(track, param_server.AddChild(
            "agent{}".format(agent_id)), start_time, end_time)
    else:
        behavior = model_converter.convert_model(behavior_model, param_server)
    try:
        initial_state = init_state_from_track(track, start_time)
    except:
        raise ValueError("Could not retrieve initial state of agent {} at t={}.".format(
            agent_id, start_time))
    bark_agent = Agent(
        initial_state,
        behavior,
        model_converter.convert_model(
            track_params["dynamic_model"], param_server),
        model_converter.convert_model(
            track_params["execution_model"], param_server),
        shape_from_track(track, param_server["DynamicModel"]["wheel_base",
                                                             "Distance between front and rear wheel center", 2.7]),
        param_server.AddChild("agent{}".format(agent_id)),
        goal_definition_from_track(track, end_time),
        track_params["map_interface"])
    # set agent id from track
    bark_agent.SetAgentId(track_id)
    return bark_agent



# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
import pandas as pd
import logging
import shutil
logging.getLogger().setLevel(logging.INFO)
import pickle
import zipfile
import uuid

from bark.runtime.scenario.scenario_generation.scenario_generation import ScenarioGeneration
from bark.runtime.commons.parameters import ParameterServer
from serialization.scenario_set_serializer import ScenarioSetSerializer

FILE_EXTENSION_SCENARIO_SET = "bark_scenarios"

# The DatabaseSerializer recursively serializes all scenario param files sets
# within a folder and releases the whole database as zip file to github


class BenchmarkDatabase:
    def __init__(self, database_root=None, dataframe=None):
        if database_root and not isinstance(dataframe, pd.DataFrame):
          self._init_from_database_root(database_root)
        elif  isinstance(dataframe, pd.DataFrame) and database_root:
          self._init_from_dataframe(dataframe, database_root)
        else:
          raise ValueError("Invalid argument combination \
                        for initialization of database")
    
    def _init_from_database_root(self, database_root):
        self.database_root = database_root
        if not os.path.exists(database_root):
            logging.error("Given database root does not exist")
            return
        if database_root.endswith("zip"):
            tmp_dir_name = "/tmp/bark_extracted_databases/{}".format(uuid.uuid4())
            logging.info("extracting zipped-database {} to temporary directory {}".format(database_root, tmp_dir_name))
            os.makedirs(tmp_dir_name)
            with zipfile.ZipFile(database_root, 'r') as zip_obj:
                zip_obj.extractall(tmp_dir_name)
            self.database_root = tmp_dir_name

        # parse recursively all info dictionaries in database into pandas table
        self.dataframe = pd.DataFrame()
        for root, dirs, files in os.walk(self.database_root):
            for file in files:
                if ScenarioSetSerializer.scenario_set_info_fileprefix() in file:
                    logging.info("Found info dict {}".format(file))
                    with open(os.path.join(root,file), "rb") as f:
                        info_dict = pickle.load(f)
                    self.dataframe = self.dataframe.append(info_dict, ignore_index=True)
        logging.info("The following scenario sets are available")
        logging.info("\n"+self.dataframe.to_string())
    
    def _init_from_dataframe(self, dataframe, database_root):
        self.database_root = database_root
        self.dataframe = dataframe

    def get_num_scenario_sets(self):
        return len(self.dataframe.index)

    def get_scenario_generator(self, scenario_set_id):
        serialized_file_name = self.dataframe.iloc[scenario_set_id]["Serialized"]
        if os.path.exists(serialized_file_name):
            serialized_file_path = serialized_file_name
        else:
            serialized_file_path = os.path.join(self.database_root, serialized_file_name)
        if os.path.exists(self.database_root):
            # move into database root that map files can be found
            os.chdir(self.database_root)
        param_file_name = self.dataframe.iloc[scenario_set_id]["Params"]
        if not param_file_name:
            logging.warning("No param file found for scenario set {}. Using defaults...".format(
                self.dataframe.iloc[scenario_set_id]["SetName"]))
            params = ParameterServer()
        else:
            params=ParameterServer(filename=param_file_name)
        scenario_generation = ScenarioGeneration(params=params)
        scenario_generation.load_scenario_list(filename=serialized_file_name)
        scenario_set_name = self.dataframe.iloc[scenario_set_id]["SetName"]
        return scenario_generation, scenario_set_name


    def __iter__(self):
        self.current_iter_idx=0
        # An iterator interface to loop over all contained scenario sets
        return self

    def __next__(self):
        if self.current_iter_idx < self.get_num_scenario_sets():
            scenario_generator = self.get_scenario_generator(self.current_iter_idx)
            self.current_iter_idx += 1
            return scenario_generator
        else:
            raise StopIteration

    def apply_filter(self, pattern, **kwargs):
        dataframe = self.dataframe[self.dataframe['SetName'].str.contains(pat=pattern, **kwargs)]
        return BenchmarkDatabase(database_root=self.database_root,
                                dataframe=dataframe)





# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import os

from load.benchmark_database import BenchmarkDatabase
from bark.benchmark.benchmark_runner import BenchmarkRunner
from bark.core.world.evaluation import EvaluatorGoalReached, \
  EvaluatorCollisionEgoAgent, EvaluatorStepCount
from bark.runtime.commons.parameters import ParameterServer

behavior_used = None
try:
  from bark.core.models.behavior import BehaviorUCTSingleAgent
  behavior_used = BehaviorUCTSingleAgent
except:
  print("BehaviorUCTSingleAgent not available, rerun example with `bazel run //bark/examples:benchmark_database --define planner_uct=true")
  exit()

db = BenchmarkDatabase(database_root="external/benchmark_database_release")
evaluators = {"success" : EvaluatorGoalReached,
              "collision" : EvaluatorCollisionEgoAgent,
              "max_steps": EvaluatorStepCount}

terminal_when = {"collision" :lambda x: x,
                 "max_steps": lambda x : x>2}

scenario_param_file ="uct_planner.json" # must be within examples params folder
params = ParameterServer(filename= os.path.join("examples/params/", scenario_param_file))
behaviors_tested = {"search5s": behavior_used(params)}

benchmark_runner = BenchmarkRunner(benchmark_database=db,
                                    evaluators=evaluators,
                                    terminal_when=terminal_when,
                                    behaviors=behaviors_tested)

benchmark_runner.run(1)
benchmark_runner.dataframe.to_pickle("uct_planner_results.pickle")



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
import numpy as np

from scipy.special import fresnel
from bark.core.world.opendrive import PlanView, XodrLaneOffset, XodrLaneWidth,\
  XodrLane, XodrRoad, FresnelCos, FresnelSin
from bark.core.geometry import Point2d


class EnvironmentTests(unittest.TestCase):
  def test_line(self):
    pv = PlanView()

    # Line
    pv.AddLine(Point2d(0, 0), 1.57079632679, 10)
    line = pv.GetReferenceLine().ToArray()

    # Spiral
    p = Point2d(line[-1][0], line[-1][1])
    pv.AddSpiral(p, 1.57079632679, 50.0, 0.0, 0.3, 0.4)
    line = pv.GetReferenceLine().ToArray()

    offset = XodrLaneOffset(1.5, 0, 0, 0)
    lane_width = XodrLaneWidth(0.0, 59.9, offset)

    lane = XodrLane.CreateLaneFromLaneWidth(-1,
                                            pv.GetReferenceLine(), lane_width, 0.5)

    print(lane)
    lane = XodrLane.CreateLaneFromLaneWidth(
        1, pv.GetReferenceLine(), lane_width, 0.5)
    print(lane)

  def test_road(self):
    newXodrRoad = XodrRoad()
    newXodrRoad.id = 1
    newXodrRoad.name = "Autobahn A9"

    newPlanView = PlanView()
    newPlanView.AddLine(Point2d(0, 0), 1.57079632679, 10)

    newXodrRoad.plan_view = newPlanView

    line = newXodrRoad.plan_view.GetReferenceLine().ToArray()

    # Spiral
    p = Point2d(line[-1][0], line[-1][1])
    newXodrRoad.plan_view.AddSpiral(p, 1.57079632679, 50.0, 0.0, 0.3, 0.4)
    line = newXodrRoad.plan_view.GetReferenceLine().ToArray()

  def test_spiral(self):
    '''
    spiral test compares outcome of wrapped odrSpiral implementation with 
    scipy fresnel calculation
    '''

    # spiral using scipy
    t = np.linspace(-7, 7, 250)
    y, x = fresnel(t)

    # spiral using odrSpiral
    x_odr = []
    y_odr = []
    for t_i in t:
      x_odr.append(FresnelCos(t_i))
      y_odr.append(FresnelSin(t_i))

    x_odr_np = np.asarray(x_odr)
    y_odr_np = np.asarray(y_odr)

    assert (np.allclose(x, x_odr_np) == 1)
    assert (np.allclose(y, y_odr_np) == 1)


if __name__ == '__main__':
  unittest.main()



# Copyright (c) 2019 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
import logging
import uuid
import subprocess
logging.getLogger().setLevel(logging.INFO)
import argparse
import zipfile
import shutil

from bark.runtime.commons.parameters import ParameterServer
from serialization.scenario_set_serializer import ScenarioSetSerializer
FILE_EXTENSION_SCENARIO_SET = "bark_scenarios"

# The DatabaseSerializer recursively serializes all scenario param files sets
# within a folder and releases the whole database as zip file to github


class DatabaseSerializer:
    def __init__(self, test_scenarios, test_world_steps, num_serialize_scenarios=None, test_scenario_idxs=None,
                                                               visualize_tests=False, viewer=None):
        self._test_scenarios = test_scenarios
        self._test_world_steps = test_world_steps
        self._test_scenario_idxs = test_scenario_idxs
        self._visualize_tests = visualize_tests
        self._database_dir = None
        self._viewer = viewer

        # provides a way to reduce the generated scenarios for each param file for unittesting
        self._num_serialize_scenarios = num_serialize_scenarios 

    def _process_folder(self, databasedir):
        process_result = True
        for root, dirs, files in os.walk(databasedir):
            for name in files:
                if name.endswith(".json"):
                    process_result = process_result and \
                           self._process_json_paramfile(os.path.join(root, name))
            for sub_dir in dirs:
                process_result = process_result and self._process_folder(sub_dir)
        return process_result

    def _process_json_paramfile(self, param_filename, json=None):
        if json:
            param_server = ParameterServer(json = json)
        else:
            param_server = ParameterServer(filename = param_filename)
        if self._num_serialize_scenarios:
            # set this down to reduce test runtime !Only for Unittesting!
            param_server["Scenario"]["Generation"]["NumScenarios"] = self._num_serialize_scenarios 
        current_dir = os.getcwd()
        os.chdir(self._database_dir)
        scenario_set_serializer = ScenarioSetSerializer(params=param_server)
        scenario_set_serializer.dump(os.path.relpath(os.path.dirname(param_filename), self._database_dir))
        scenario_set_serializer.load()
        test_result = scenario_set_serializer.test(num_scenarios=self._test_scenarios,
                                     num_steps=self._test_world_steps,
                                     visualize_test=self._visualize_tests,
                                     viewer=self._viewer,
                                     test_scenario_idxs=self._test_scenario_idxs)
        os.chdir(current_dir)
        return test_result
    
    def _process_scenario_list(self, database_dir, scenario_set_dict):
        serialized_sets_dir = os.path.join(database_dir, "scenario_sets")
        if not os.path.exists(serialized_sets_dir):
            os.makedirs(serialized_sets_dir)
        process_result = True
        for scenario_set, json_params in scenario_set_dict.items():
            json_params["Scenario"]["Generation"]["SetName"] = scenario_set
            params = ParameterServer(json=json_params)
            filename = os.path.join(serialized_sets_dir, "{}.json".format(scenario_set))
            print(filename)
            params.save(filename)
            process_result = process_result and \
                    self._process_json_paramfile(filename, json=json_params)
        return process_result

    def process(self, database_dir, scenario_set_dict=None):
        if not scenario_set_dict:
            self._database_dir = database_dir
            succ = self._process_folder(database_dir)
        elif database_dir and scenario_set_dict:
            self._database_dir = database_dir # TODO(@Klemens): use serialized_sets_dir?
            succ = self._process_scenario_list(database_dir, scenario_set_dict)
        else:
            raise ValueError("Invalid argument combination.")
        return succ

    @staticmethod
    def _release_file_name(version):
        return "benchmark_database_{}".format(version)

    @staticmethod
    def _pack(database_dir, packed_file_name):
        logging.info('The following list of files will be released:')
        zipf = zipfile.ZipFile(packed_file_name, 'w', zipfile.ZIP_DEFLATED)
        tmp_dir = "/tmp/bark_packed_databases/{}".format(uuid.uuid4())
        shutil.copytree(database_dir, tmp_dir) # copy to resolve symlinks
        for subdir, dirs, files in os.walk(tmp_dir):
            for file in files:
                filepath = os.path.join(subdir, file)
                if filepath.endswith(".zip"):
                    continue # do not include our own created zip file or BUILD
                zip_root = filepath.replace(tmp_dir,"")
                logging.info(os.path.join(zip_root, file))
                zipf.write(os.path.join(subdir, file), zip_root)
        logging.info("Packed release file {}".format(os.path.abspath(packed_file_name)))

    @staticmethod
    def _github_release(database_file, version, github_token, delete):
        if delete:
            delete_cmd = "-delete"
        else:
            delete_cmd = ""
        print(os.getcwd())
        ghr_command = os.path.abspath("external/ghr")
        full_file_path = os.path.abspath(database_file)
        os.chdir(ghr_command)
        ghr_full_command = "./ghr -t {} -u bark-simulator -r benchmark-database {} {} {}".format(
           github_token, delete_cmd, version, full_file_path)
        out = subprocess.call(ghr_full_command, shell=True)
        logging.info("Release output {}".format(out))

    def release(self, version, github_token=None, delete=False):
        if self._num_serialize_scenarios:
            logging.warning("Do not set down the number of serialized scenarios in a release \
                    -> set serialized_scenarios=None")

        if not self._database_dir:
            logging.error("No database dir given, call process first.")

        packed_file_name = os.path.join(os.path.dirname(self._database_dir),"{}.{}".format(
            DatabaseSerializer._release_file_name(version), "zip"))
        DatabaseSerializer._pack(self._database_dir, packed_file_name)

        if github_token:
            DatabaseSerializer._github_release(packed_file_name, version, github_token, delete)
        else:
            logging.info("Assuming local release as you did not provide a github token.")

        return packed_file_name




import unittest
import pickle
import numpy as np

from bark.core.world.agent import *
from bark.core.models.behavior import *
from bark.core.world import *
from bark.core.world.map import *
from bark.core.geometry import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *
from bark.core.geometry import *
from bark.core.geometry.standard_shapes import *
from bark.core.world.goal_definition import *
from bark.runtime.commons.parameters import ParameterServer

def pickle_unpickle(object):
    with open('temp.pickle','wb') as f:
        pickle.dump(object,f)
    object = None
    with open( 'temp.pickle', "rb" ) as f:
        object = pickle.load(f)
    return object


class PickleTests(unittest.TestCase):

    def test_geometry_pickle(self):
        # point 2d
        p = Point2d(2 ,3)

        pa = pickle_unpickle(p)
        self.assertEqual(p.x(), pa.x())
        self.assertEqual(p.y(), pa.y())

        # linestring
        l = Line2d()
        l.AddPoint(p)
        l.AddPoint(Point2d(10,4))
        l.AddPoint(Point2d(1.555555, 1.244222))

        la = pickle_unpickle(l)
        self.assertTrue(np.array_equal(l.ToArray(), la.ToArray()))

        l = Line2d()
        la = pickle_unpickle(l)
        self.assertTrue(np.array_equal(l.ToArray(), la.ToArray()))

        # polygon
        p = CarLimousine()
        pa = pickle_unpickle(p)
        self.assertTrue(np.array_equal(p.ToArray(), pa.ToArray()))

        # polygon
        pr = CarRectangle()
        pra = pickle_unpickle(pr)
        self.assertTrue(np.array_equal(pr.ToArray(), pra.ToArray()))

    def test_behavior_model_pickle(self):
        
        params = ParameterServer()
        b = BehaviorConstantVelocity(params)

        ba = pickle_unpickle(b)
        self.assertTrue(isinstance(ba, BehaviorConstantVelocity))

    def test_execution_model_pickle(self):
        
        params = ParameterServer()
        e = ExecutionModelInterpolate(params)

        ea = pickle_unpickle(e)
        self.assertTrue(isinstance(ea,ExecutionModelInterpolate))

    def test_dynamic_model_pickle(self):
        params = ParameterServer()
        d = SingleTrackModel(params)

        da = pickle_unpickle(d)
        self.assertTrue(isinstance(da, SingleTrackModel))


    def test_goal_definition(self):
        goal_polygon = Polygon2d([0, 0, 0],[Point2d(-1,-1),Point2d(-1,1),Point2d(1,1), Point2d(1,-1)])
        goal_definition = GoalDefinitionStateLimits(goal_polygon, (0.2 , 0.5))

        goal_definition_after = pickle_unpickle(goal_definition)

        self.assertTrue(np.array_equal(goal_definition.xy_limits.center, \
                                        goal_definition_after.xy_limits.center))


        goal_definition2 = GoalDefinitionStateLimits(goal_polygon, (0.2 , 0.5))
        goal_definition_sequential = GoalDefinitionSequential([goal_definition, goal_definition2])
        goal_definition_sequential_after = pickle_unpickle(goal_definition_sequential)

        sequential_goals_after = goal_definition_sequential_after.sequential_goals
        self.assertTrue(np.array_equal(sequential_goals_after[0].xy_limits.ToArray(), \
                                        goal_definition.xy_limits.ToArray()))

        self.assertTrue(np.array_equal(sequential_goals_after[1].xy_limits.ToArray(), \
                                        goal_definition2.xy_limits.ToArray()))

    def test_agent_pickle(self):
        params = ParameterServer()
        behavior = BehaviorConstantVelocity(params)
        execution = ExecutionModelInterpolate(params)
        dynamic = SingleTrackModel(params)
        shape = CarLimousine()
        init_state = np.array([0, 0, 0, 0, 5])
        goal_polygon = Polygon2d([0, 0, 0],[Point2d(-1,-1),Point2d(-1,1),Point2d(1,1), Point2d(1,-1)])
        goal_definition = GoalDefinitionPolygon(goal_polygon)
        agent = Agent(init_state, behavior, dynamic, execution, shape, params.AddChild("agent"), goal_definition )

        agent_after = pickle_unpickle(agent)

        self.assertEqual(agent_after.id , agent.id)
        self.assertTrue(np.array_equal(agent_after.state, agent.state) )
        self.assertTrue(np.array_equal(agent_after.goal_definition.goal_shape.center, \
                                       agent.goal_definition.goal_shape.center))

        goal_definition_2 = GoalDefinitionStateLimits(goal_polygon, (0.2 , 0.5))
        agent2 = Agent(init_state, behavior, dynamic, execution, shape, params.AddChild("agent"), goal_definition_2)

        agent_after2 = pickle_unpickle(agent2)

        self.assertEqual(agent_after2.id , agent2.id)
        self.assertTrue(np.array_equal(agent_after2.state, agent.state) )
        self.assertTrue(np.array_equal(agent_after2.goal_definition.xy_limits.center, \
                                       agent2.goal_definition.xy_limits.center))

        agent_list = []
        agent_list.append(agent)

        agent_list_after = pickle_unpickle(agent_list)

        self.assertEqual(agent_list_after[0].id , agent.id)
        self.assertTrue(np.array_equal(agent_list_after[0].state, agent.state) )

    def test_agent_pickle_uct_planner(self):
        try:
            from bark.core.models.behavior import BehaviorUCTSingleAgentMacroActions
        except:
            print("Rerun test with ---define planner_uct=true")
            return

        params = ParameterServer()
        behavior = BehaviorUCTSingleAgentMacroActions(params)
        execution = ExecutionModelInterpolate(params)
        dynamic = SingleTrackModel(params)
        shape = CarLimousine()
        init_state = np.array([0, 0, 0, 0, 5])
        goal_polygon = Polygon2d([0, 0, 0],[Point2d(-1,-1),Point2d(-1,1),Point2d(1,1), Point2d(1,-1)])
        goal_definition = GoalDefinitionPolygon(goal_polygon)
        agent = Agent(init_state, behavior, dynamic, execution, shape, params.AddChild("agent"), goal_definition )

        agent_after = pickle_unpickle(agent)

        self.assertTrue(isinstance(agent_after.behavior_model, 
                                BehaviorUCTSingleAgentMacroActions))



if __name__ == '__main__':
    unittest.main()


# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np

from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces \
       import ConfigReaderAgentStatesAndGeometries

from bark.core.geometry.standard_shapes import *
from bark.core.geometry import *

# this config reader defines agent states with distances sampled uniformly standard vehicle geometries models
# it can be specified with parameter "lane_position" being between 1 and num_road_corridor_lanes 
# in what lanes vehicles are placed, or if None in all lanes they are placed
class UniformVehicleDistribution(ConfigReaderAgentStatesAndGeometries):
  def create_from_config(self, config_param_object, road_corridor):
    self._lane_positions = config_param_object["LanePositions", "List of values out of 0:max_num_lanes-1. \
            Vehicles are placed only this lanes \
           or None if vehicles shall be placed at alles lanes", None]
    self._vehicle_distance_range = config_param_object["VehicleDistanceRange",
      "Distance range between vehicles in meter given as tuple from which" + \
      "distances are sampled uniformly",
      (10, 20)]
    self._other_velocity_range = config_param_object["OtherVehicleVelocityRange",
      "Lower and upper bound of velocity in km/h given as tuple from which" + \
      " velocities are sampled uniformly",
      (20,30)]
    self._s_range = config_param_object["SRange", "pair of values (tuple) between 0 and 1 to limit placement  \
                     to certain part of road corridor", (0.1, 0.7)]
    # todo make parameterizable, but not only based on 2d points
    self._vehicle_2d_shape = CarLimousine()

    agent_states = []
    agent_geometries = []
    agent_lane_positions = []

    lane_corridors, lane_positions = self.select_lane_corridors(road_corridor, self._lane_positions)
    for idx, lane_corridor in enumerate(lane_corridors):
      tmp_agent_states, tmp_agent_geometries = self.agents_along_lane_corridor(lane_corridor,
                   self._s_range[0], self._s_range[1])

      agent_states.extend(tmp_agent_states)
      agent_geometries.extend(tmp_agent_geometries)
      agent_lane_positions.extend([lane_positions[idx]]*len(tmp_agent_states))

    assert(len(agent_states) == len(agent_geometries))
    assert(len(agent_states) == len(agent_lane_positions))
    return agent_states, agent_geometries, {"agent_lane_positions": agent_lane_positions}, config_param_object


  def select_lane_corridors(self, road_corridor, lane_positions):
    lane_corridors = []
    if lane_positions:
      for lane_position in lane_positions:
        lane_corridors.append(road_corridor.lane_corridors[lane_position])
    else:
        lane_corridors = road_corridor.lane_corridors
        lane_positions = list(range(1, len(lane_corridors)+1))
    return lane_corridors, lane_positions

  def sample_velocity_uniform(self, velocity_range):
    return self.random_state.uniform(velocity_range[0], velocity_range[1])

  def sample_distance_uniform(self, distance_range):
    return self.random_state.uniform(distance_range[0], distance_range[1])


  def agents_along_lane_corridor(self,
                              lane_corridor,
                              s_start,
                              s_end):
    linestring = lane_corridor.center_line
    agent_states = []
    agent_geometries = []
    corridor_length = lane_corridor.center_line.Length()
    s = s_start*corridor_length
    while s < s_end*corridor_length:
      # set agent state on linestring with random velocity
      xy_point =  GetPointAtS(linestring, s)
      angle = GetTangentAngleAtS(linestring, s)
      
      velocity = self.sample_velocity_uniform(self._other_velocity_range)
      agent_state = [0, xy_point.x(), xy_point.y(), angle, velocity ]

      agent_states.append(agent_state)
      agent_geometries.append(self._vehicle_2d_shape)

      # move forward on linestring based on vehicle size and max/min distance
      s += self._vehicle_2d_shape.front_dist + self._vehicle_2d_shape.rear_dist + \
                  self.sample_distance_uniform(self._vehicle_distance_range)
    return agent_states, agent_geometries




# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import math
from bark.core.geometry.standard_shapes import *
from bark.core.geometry import *
import unittest
import os
import matplotlib as mpl
if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')
from bark.core.geometry import *
from bark.core.geometry.standard_shapes import *
import math
import numpy as np

#from bark.visualization import MPViewer


class GeometryTests(unittest.TestCase):
    def test_init_objects(self):
        p = Point2d(0.0, 1.0)
        print(p)
        print(p.x)
        print(p.y)

        l = Line2d()
        l.AddPoint(p)
        l.AddPoint(p)
        l.AddPoint(p)
        self.assertFalse(l.Valid())
        print(l)

        p = Point2d(0.0, 1.0)
        l.AddPoint([3, 4])
        l.AddPoint([1, 4])
        self.assertTrue(l.Valid())
        arr = l.ToArray()
        print(arr)

        p = Polygon2d()
        p.AddPoint([0, 0])
        p.AddPoint([1, 0])
        p.AddPoint([1, 1])
        # to few points
        self.assertFalse(p.Valid())
        p.AddPoint([0, 0])
        # wrong orientation
        self.assertFalse(p.Valid())

        p = Polygon2d()
        p.AddPoint([0, 0])
        p.AddPoint([0, 1])
        p.AddPoint([1, 1])
        p.AddPoint([0, 0])
        # right orientation
        self.assertTrue(p.Valid())

    def test_bounding_box(self):
        p = Polygon2d()
        p.AddPoint([0, 0])
        p.AddPoint([0, 1])
        p.AddPoint([1, 1])
        p.AddPoint([0, 0])
        # right orientation
        self.assertTrue(p.Valid())

        bb = p.bounding_box
        self.assertTrue(np.array_equal(
            np.array([0, 0]), np.array([bb[0].x(), bb[0].y()])))
        self.assertTrue(np.array_equal(
            np.array([1, 1]), np.array([bb[1].x(), bb[1].y()])))

        p = Polygon2d()
        p.AddPoint([2, 3])
        p.AddPoint([2, 4])
        p.AddPoint([3, 4])
        p.AddPoint([2, 3])
        # right orientation
        self.assertTrue(p.Valid())

        bb2 = p.bounding_box
        self.assertTrue(np.array_equal(
            np.array([2, 3]), np.array([bb2[0].x(), bb2[0].y()])))
        self.assertTrue(np.array_equal(
            np.array([3, 4]), np.array([bb2[1].x(), bb2[1].y()])))

        bb_merged = MergeBoundingBoxes(bb, bb2)

        self.assertTrue(np.array_equal(np.array([0, 0]), np.array(
            [bb_merged[0].x(), bb_merged[0].y()])))
        self.assertTrue(np.array_equal(np.array([3, 4]), np.array(
            [bb_merged[1].x(), bb_merged[1].y()])))

    def test_Distance(self):

        p1 = Point2d(1, 10)
        p2 = Point2d(1, 5)

        self.assertEqual(Distance(p1, p2), 5)

        l = Line2d()
        l.AddPoint(p1)
        l.AddPoint(p2)

        p3 = Point2d(4, 15)
        p4 = Point2d(4, -5)

        l2 = Line2d()
        l2.AddPoint(p3)
        l2.AddPoint(p4)

        self.assertEqual(Distance(l2, p3), 0)
        self.assertEqual(Distance(l, l2), 3)

        p = Polygon2d()
        p.AddPoint([0, 0])
        p.AddPoint([0, 1])
        p.AddPoint([1, 1])
        p.AddPoint([0, 0])

        p2 = Polygon2d()
        p2.AddPoint([5, 0])
        p2.AddPoint([5, 1])
        p2.AddPoint([6, 1])
        p2.AddPoint([5, 0])

        self.assertEqual(Distance(p, p2), 4)
        self.assertEqual(Distance(p, l2), 3)
        self.assertEqual(Distance(p2, Point2d(10, 1)), 4)

    # TODO: need assertion in this test
    def test_transformation(self):
        #viewer = MPViewer()

        p = Polygon2d()
        p.AddPoint([0, 0])
        p.AddPoint([0, 2])
        p.AddPoint([4, 2])
        p.AddPoint([4, 0])
        p.AddPoint([0, 0])
        self.assertTrue(p.Valid())

        # rotate around pi
        # viewer.drawPolygon2d(p.Rotate(math.pi/2),'r',0.2)
        p.Rotate(math.pi / 2)

        # rotate around top left corner (second point)
        p2 = Polygon2d([1.25, 1, 0], [
            Point2d(0, 0),
            Point2d(0, 2),
            Point2d(4, 2),
            Point2d(4, 0),
            Point2d(0, 0)
        ])
        for x in np.linspace(0, 2 * math.pi, 7):
            p2.Rotate(x)
            #viewer.drawPolygon2d(p2.Rotate(x), 'b', 0.2)

        # translate into the nirvana based on previous rotations and translations
        #viewer.drawPolygon2d(p2.Translate(Point2d(10,0)), 'g', 0.2)
        #viewer.drawPolygon2d(p2.Transform([5,5,0.4]), 'k', 0.2)
        # viewer.show()

    def test_const_poly_from_array(self):
        arr = np.array([[0, 0], [0, 2], [4, 2], [4, 0], [0, 0]])
        poly = Polygon2d([1, 3, 1], arr)
        self.assertTrue(poly.Valid())

    def test_concatenate_polygons(self):

        poly1 = Polygon2d()
        poly1.AddPoint([0, 0])
        poly1.AddPoint([0, 2])
        poly1.AddPoint([4, 2])
        poly1.AddPoint([4, 0])
        poly1.AddPoint([0, 0])
        self.assertTrue(poly1.Valid())
        a1 = poly1.CalculateArea()

        poly2 = Polygon2d()
        poly2.AddPoint([4, 0])
        poly2.AddPoint([4, 2])
        poly2.AddPoint([8, 2])
        poly2.AddPoint([8, 0])
        poly2.AddPoint([4, 0])
        self.assertTrue(poly2.Valid())
        a2 = poly2.CalculateArea()

        poly1.ConcatenatePolygons(poly2)
        self.assertTrue(poly1.Valid())

        a3 = poly1.CalculateArea()

        self.assertTrue(a3 == a1+a2)

    def test_robust_concatenate_polygons(self):

        poly1 = Polygon2d()
        poly1.AddPoint([0, 0])
        poly1.AddPoint([0, 2])
        poly1.AddPoint([4, 2])
        poly1.AddPoint([4, 0])
        poly1.AddPoint([0, 0])
        self.assertTrue(poly1.Valid())
        a1 = poly1.CalculateArea()

        eps = 0.05
        poly2 = Polygon2d()
        poly2.AddPoint([4+eps, 0])
        poly2.AddPoint([4+eps, 2])
        poly2.AddPoint([8+eps, 2])
        poly2.AddPoint([8+eps, 0])
        poly2.AddPoint([4+eps, 0])
        self.assertTrue(poly2.Valid())
        a2 = poly2.CalculateArea()

        poly1b = poly1.BufferPolygon(0.1)
        poly1b.ConcatenatePolygons(poly2)
        poly3 = poly1b.BufferPolygon(-0.1)
        self.assertTrue(poly3.Valid())

        a3 = poly3.CalculateArea()
        print(poly3)

    def test_distances_to_center(self):

        shape = CarLimousine()

        self.assertAlmostEqual(shape.front_dist, 3.849999, places=4)
        self.assertAlmostEqual(shape.rear_dist, 1.118999, places=4)
        self.assertAlmostEqual(shape.left_dist, 0.955999, places=4)
        self.assertAlmostEqual(shape.right_dist, 0.955999, places=4)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
from unittest.mock import patch
import time
import math
import itertools
import filecmp
import os
import matplotlib.pyplot as plt

from bark.core.world import World
from bark.core.geometry import *
from bark.runtime.commons.parameters import ParameterServer
from bark.core.world.opendrive import XodrDrivingDirection, MakeXodrMapOneRoadTwoLanes
from bark.core.world.map import MapInterface

from bark.runtime.commons import XodrParser
from bark.runtime.viewer import MPViewer
import numpy as np


class RoadCorridorTests(unittest.TestCase):
    @unittest.skip("...")
    def test_road_corridor_forward(self):
        xodr_parser = XodrParser(os.path.join(os.path.dirname(__file__), \
            "../../runtime/tests/data/road_corridor_test.xodr"))

        # World Definition
        params = ParameterServer()
        world = World(params)

        map_interface = MapInterface()
        map_interface.SetOpenDriveMap(xodr_parser.map)
        world.SetMap(map_interface)
        open_drive_map = world.map.GetOpenDriveMap()
        viewer = MPViewer(params=params,
                          use_world_bounds=True)

        # Draw map
        viewer.drawWorld(world)
        viewer.show(block=False)

        # Generate RoadCorridor
        roads = [0, 1, 2]
        driving_direction = XodrDrivingDirection.forward
        map_interface.GenerateRoadCorridor(roads, driving_direction)
        road_corridor = map_interface.GetRoadCorridor(roads, driving_direction)

        # Assert road corridor

        # Assert: 3 roads
        self.assertEqual(len(road_corridor.roads), 3)

        # Assert: road1: 2 lanes, road2: 1 lane, road3: 1 lane
        self.assertEqual(len(road_corridor.GetRoad(0).lanes), 3)
        self.assertEqual(len(road_corridor.GetRoad(1).lanes), 2)
        self.assertEqual(len(road_corridor.GetRoad(2).lanes), 3)

        # Assert: next road
        self.assertEqual(road_corridor.GetRoad(0).next_road.road_id, 1)
        self.assertEqual(road_corridor.GetRoad(1).next_road.road_id, 2)

        # Assert: lane links
        self.assertEqual(road_corridor.GetRoad(
            0).GetLane(3).next_lane.lane_id, 5)
        self.assertEqual(road_corridor.GetRoad(
            1).GetLane(5).next_lane.lane_id, 8)

        # Assert: LaneCorridor
        self.assertEqual(len(road_corridor.lane_corridors), 3)

        colors = ["blue", "red", "green"]
        count = 0
        for lane_corridor in road_corridor.lane_corridors:
            viewer.drawPolygon2d(lane_corridor.polygon,
                                 color=colors[count], alpha=0.5)
            viewer.drawLine2d(lane_corridor.left_boundary, color="red")
            viewer.drawLine2d(lane_corridor.right_boundary, color="blue")
            viewer.drawLine2d(lane_corridor.center_line, color="black")
            viewer.show(block=False)
            plt.pause(2.)
            count += 1
    
    def test_three_way_intersection(self):
        # threeway_intersection
        xodr_parser = XodrParser(os.path.join(os.path.dirname(__file__), \
            "../../runtime/tests/data/threeway_intersection.xodr"))

        # World Definition
        params = ParameterServer()
        world = World(params)

        map_interface = MapInterface()
        map_interface.SetOpenDriveMap(xodr_parser.map)
        world.SetMap(map_interface)
        open_drive_map = world.map.GetOpenDriveMap()
        viewer = MPViewer(params=params,
                          use_world_bounds=True)
        comb_all = []
        start_point = [Point2d(-30, -2)]
        end_point_list = [Point2d(30, -2), Point2d(-2, -30)]
        comb = list(itertools.product(start_point, end_point_list))
        comb_all = comb_all + comb

        # starting on the right
        start_point = [Point2d(30, 2)]
        end_point_list = [Point2d(-30, 2)]
        comb = list(itertools.product(start_point, end_point_list))
        comb_all = comb_all + comb

        # starting on the bottom
        start_point = [Point2d(2, -30)]
        end_point_list = [Point2d(30, -2), Point2d(-30, 2)]
        comb = list(itertools.product(start_point, end_point_list))
        comb_all = comb_all + comb

        # check few corridors
        def GenerateRoadCorridor(map_interface, comb):
            (start_p, end_p) = comb
            polygon = Polygon2d(
                [0, 0, 0], [Point2d(-1, -1), Point2d(-1, 1), Point2d(1, 1), Point2d(1, -1)])
            start_polygon = polygon.Translate(start_p)
            goal_polygon = polygon.Translate(end_p)
            rc = map_interface.GenerateRoadCorridor(start_p, goal_polygon)
            return rc

        # assert road ids
        rc = GenerateRoadCorridor(map_interface, comb_all[0])
        self.assertEqual(rc.road_ids, [0, 11, 1])
        self.assertEqual(len(rc.lane_corridors), 3)
        self.assertTrue(rc.polygon.Valid())
        rc = GenerateRoadCorridor(map_interface, comb_all[1])
        self.assertEqual(rc.road_ids, [0, 5, 2])
        self.assertEqual(len(rc.lane_corridors), 3)
        self.assertTrue(rc.polygon.Valid())
        rc = GenerateRoadCorridor(map_interface, comb_all[2])
        self.assertEqual(rc.road_ids, [1, 10, 0])
        self.assertEqual(len(rc.lane_corridors), 3)
        self.assertTrue(rc.polygon.Valid())
        rc = GenerateRoadCorridor(map_interface, comb_all[3])
        self.assertEqual(rc.road_ids, [2, 6, 1])
        self.assertEqual(len(rc.lane_corridors), 3)
        self.assertTrue(rc.polygon.Valid())
        rc = GenerateRoadCorridor(map_interface, comb_all[4])
        self.assertEqual(rc.road_ids, [2, 4, 0])
        self.assertEqual(len(rc.lane_corridors), 3)
        self.assertTrue(rc.polygon.Valid())


    def test_dr_deu_merging(self):
        # threeway_intersection
        xodr_parser = XodrParser(os.path.join(os.path.dirname(__file__), \
            "../../runtime/tests/data/DR_DEU_Merging_MT_v01_shifted.xodr"))

        # World Definition
        params = ParameterServer()
        world = World(params)

        map_interface = MapInterface()
        map_interface.SetOpenDriveMap(xodr_parser.map)
        world.SetMap(map_interface)

        roads = [0, 1]
        driving_direction = XodrDrivingDirection.forward
        map_interface.GenerateRoadCorridor(roads, driving_direction)
        road_corridor = map_interface.GetRoadCorridor(roads, driving_direction)

        # Draw map
        viewer = MPViewer(params=params, use_world_bounds=True)
        viewer.drawWorld(world)

        viewer.drawPolygon2d(road_corridor.lane_corridors[0].polygon, color = "blue", alpha = 0.5)
        viewer.drawPolygon2d(road_corridor.lane_corridors[1].polygon, color = "blue", alpha = 0.5)
        viewer.show(block=False)

        self.assertTrue(road_corridor.lane_corridors[0].polygon.Valid())
        self.assertTrue(road_corridor.lane_corridors[1].polygon.Valid())
        self.assertTrue(road_corridor.polygon.Valid())

    def test_dr_chn_merging(self):
        # threeway_intersection
        xodr_parser = XodrParser(os.path.join(os.path.dirname(__file__), "../../runtime/tests/data/DR_CHN_Merging_ZS_partial_v02.xodr"))

        # World Definition
        params = ParameterServer()
        world = World(params)

        map_interface = MapInterface()
        map_interface.SetOpenDriveMap(xodr_parser.map)
        world.SetMap(map_interface)

        roads = [0, 1]
        driving_direction = XodrDrivingDirection.forward
        map_interface.GenerateRoadCorridor(roads, driving_direction)
        road_corridor = map_interface.GetRoadCorridor(roads, driving_direction)

        # Draw map
        viewer = MPViewer(params=params, use_world_bounds=True)
        viewer.drawWorld(world)

        viewer.drawPolygon2d(road_corridor.lane_corridors[0].polygon, color = "blue", alpha = 0.5)
        viewer.drawPolygon2d(road_corridor.lane_corridors[1].polygon, color = "blue", alpha = 0.5)
        viewer.drawPolygon2d(road_corridor.lane_corridors[2].polygon, color = "blue", alpha = 0.5)
        viewer.show(block=False)

        self.assertTrue(road_corridor.lane_corridors[0].polygon.Valid())
        self.assertTrue(road_corridor.lane_corridors[1].polygon.Valid())
        self.assertTrue(road_corridor.lane_corridors[2].polygon.Valid())
        self.assertTrue(road_corridor.polygon.Valid())


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import os
import pickle
import pandas as pd
import logging
import copy
import time

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)

from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.scenario.scenario import Scenario
from bark.core.world.evaluation import *


# contains information for a single benchmark run
class BenchmarkConfig:
    def __init__(self, config_idx, behavior, behavior_name,
                 scenario, scenario_idx, scenario_set_name):
        self.config_idx = config_idx
        self.behavior = behavior
        self.behavior_name = behavior_name
        self.scenario = scenario
        self.scenario_idx = scenario_idx
        self.scenario_set_name = scenario_set_name

    def get_info_string_list(self):
        info_strings = ["ConfigIdx: {}".format(self.config_idx),
                        "Behavior: {}".format(self.behavior_name),
                        "ScenarioSet: {}".format(self.scenario_set_name),
                        "ScenarioIdx: {}".format(self.scenario_idx)]
        return info_strings


# result of benchmark run
class BenchmarkResult:
    def __init__(self, result_dict, benchmark_configs, histories=None):
        self.__result_dict = result_dict
        self.__benchmark_configs = benchmark_configs
        self.__data_frame = None
        self.__histories = histories or []

    def get_data_frame(self):
        if not isinstance(self.__data_frame, pd.DataFrame):
            self.__data_frame = pd.DataFrame(self.__result_dict)
        return self.__data_frame

    def get_result_dict(self):
        return self.__result_dict

    def get_benchmark_configs(self):
        return self.__benchmark_configs

    def get_histories(self):
        return self.__histories

    def get_benchmark_config(self, config_idx):
        return BenchmarkResult.find_benchmark_config(
            self.__benchmark_configs, config_idx)

    def get_history(self, config_idx):
        return self.__histories[config_idx]

    @staticmethod
    def find_benchmark_config(benchmark_configs, config_idx):
        BenchmarkResult._sort_bench_confs(benchmark_configs)
        bench_conf = benchmark_configs[config_idx]
        assert (bench_conf.config_idx == config_idx)
        return bench_conf

    @staticmethod
    def _sort_bench_confs(benchmark_configs):
        def sort_key(bench_conf):
            return bench_conf.config_idx

        benchmark_configs.sort(key=sort_key)

    @staticmethod
    def load(filename):
        with open(filename, 'rb') as handle:
            dmp = pickle.load(handle)
        return dmp

    def dump(self, filename, dump_configs=True):
        if isinstance(self.__data_frame, pd.DataFrame):
            self.__data_frame = None
        if not dump_configs:
            self.__benchmark_configs = None
        with open(filename, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)
        logging.info("Saved BenchmarkResult to {}".format(
            os.path.abspath(filename)))


class BenchmarkRunner:
    def __init__(self,
                 benchmark_database=None,
                 evaluators=None,
                 terminal_when=None,
                 behaviors=None,
                 num_scenarios=None,
                 benchmark_configs=None,
                 logger_name=None,
                 log_eval_avg_every=None):

        self.benchmark_database = benchmark_database
        self.evaluators = evaluators or {}
        self.terminal_when = terminal_when or []
        self.behaviors = behaviors or {}
        self.benchmark_configs = benchmark_configs or \
                                 self._create_configurations(num_scenarios)
        self.exceptions_caught = []
        self.log_eval_avg_every = log_eval_avg_every
        self.logger = logging.getLogger(logger_name or "BenchmarkRunner")
        self.logger.setLevel(logging.DEBUG)

    def _create_configurations(self, num_scenarios=None):
        benchmark_configs = []
        for behavior_name, behavior_bark in self.behaviors.items():
            # run over all scenario generators from benchmark database
            for scenario_generator, scenario_set_name in self.benchmark_database:
                for scenario, scenario_idx in scenario_generator:
                    if num_scenarios and scenario_idx >= num_scenarios:
                        break
                    benchmark_config = \
                        BenchmarkConfig(
                            len(benchmark_configs),
                            behavior_bark,
                            behavior_name,
                            scenario,
                            scenario_idx,
                            scenario_set_name
                        )
                    benchmark_configs.append(benchmark_config)
        return benchmark_configs

    def run(self, viewer=None, maintain_history=False):
        results = []
        histories = {}
        for idx, bmark_conf in enumerate(self.benchmark_configs):
            self.logger.info("Running config idx {}/{}: Scenario {} of set \"{}\" for behavior \"{}\"".format(
                idx, len(self.benchmark_configs) - 1, bmark_conf.scenario_idx,
                bmark_conf.scenario_set_name, bmark_conf.behavior_name))
            result_dict, scenario_history = self._run_benchmark_config(copy.deepcopy(bmark_conf), viewer,
                                                                       maintain_history)
            results.append(result_dict)
            histories[bmark_conf.config_idx] = scenario_history
            if self.log_eval_avg_every and (idx + 1) % self.log_eval_avg_every == 0:
                self._log_eval_average(results)
        return BenchmarkResult(results, self.benchmark_configs, histories=histories)

    def run_benchmark_config(self, config_idx, **kwargs):
        for idx, bmark_conf in enumerate(self.benchmark_configs):
            if bmark_conf.config_idx == config_idx:
                return self._run_benchmark_config(copy.deepcopy(bmark_conf), **kwargs)
        self.logger.error("Config idx {} not found in benchmark configs. Skipping...".format(config_idx))
        return

    def _run_benchmark_config(self, benchmark_config, viewer=None, maintain_history=False):
        scenario = benchmark_config.scenario
        behavior = benchmark_config.behavior
        parameter_server = ParameterServer(json=scenario._json_params)
        scenario_history = []
        step = 0
        try:
            world = scenario.GetWorldState()
        except Exception as e:
            self.logger.error("For config-idx {}, Exception thrown in scenario.GetWorldState: {}".format(
                benchmark_config.config_idx, e))
            self._append_exception(benchmark_config, e)
            return {"config_idx": benchmark_config.config_idx,
                    "scen_set": benchmark_config.scenario_set_name,
                    "scen_idx": benchmark_config.scenario_idx,
                    "step": step,
                    "behavior": benchmark_config.behavior_name,
                    "Terminal": "exception_raised"}

        # if behavior is not None (None specifies that also the default model can be evalauted)
        if behavior:
            world.agents[scenario._eval_agent_ids[0]].behavior_model = behavior
        if maintain_history:
            self._append_to_scenario_history(scenario_history, world, scenario)
        self._reset_evaluators(world, scenario._eval_agent_ids)
        step_time = parameter_server["Simulation"]["StepTime", "", 0.2]
        if not isinstance(step_time, float):
            step_time = 0.2
        terminal = False
        terminal_why = None
        while not terminal:
            try:
                evaluation_dict = self._get_evalution_dict(world)
            except Exception as e:
                self.logger.error("For config-idx {}, Exception thrown in evaluation: {}".format(
                    benchmark_config.config_idx, e))
                terminal_why = "exception_raised"
                self._append_exception(benchmark_config, e)
                evaluation_dict = {}
                break
            terminal, terminal_why = self._is_terminal(evaluation_dict)
            if not terminal:
                if viewer:
                    viewer.drawWorld(world, scenario._eval_agent_ids, scenario_idx=benchmark_config.scenario_idx)
                    viewer.show(block=False)
                    time.sleep(step_time)
                    viewer.clear()
                try:
                    world.Step(step_time)
                except Exception as e:
                    self.logger.error("For config-idx {}, Exception thrown in world.Step: {}".format(
                        benchmark_config.config_idx, e))
                    terminal_why = "exception_raised"
                    self._append_exception(benchmark_config, e)
                    break

                if maintain_history:
                    self._append_to_scenario_history(scenario_history, world, scenario)
                step += 1

        dct = {"config_idx": benchmark_config.config_idx,
               "scen_set": benchmark_config.scenario_set_name,
               "scen_idx": benchmark_config.scenario_idx,
               "step": step,
               "behavior": benchmark_config.behavior_name,
               **evaluation_dict,
               "Terminal": terminal_why}

        return dct, scenario_history

    def _append_to_scenario_history(self, scenario_history, world, scenario):
        scenario = Scenario(agent_list=list(world.agents.values()),
                            map_file_name=scenario.map_file_name,
                            eval_agent_ids=scenario.eval_agent_ids,
                            json_params=scenario.json_params)
        scenario_history.append(scenario.copy())

    def _append_exception(self, benchmark_config, exception):
        self.exceptions_caught.append((benchmark_config.config_idx, exception))

    def _reset_evaluators(self, world, eval_agent_ids):
        for evaluator_name, evaluator_type in self.evaluators.items():
            evaluator_bark = None
            try:
                evaluator_bark = eval("{}(eval_agent_ids[0])".format(evaluator_type))
            except:
                evaluator_bark = eval("{}()".format(evaluator_type))
            world.AddEvaluator(evaluator_name, evaluator_bark)

    def _evaluation_criteria(self):
        bark_evals = [eval_crit for eval_crit, _ in self.evaluators.items()]
        bark_evals.append("step")
        return bark_evals

    def _get_evalution_dict(self, world):
        return world.Evaluate()

    def _is_terminal(self, evaluation_dict):
        terminal = False
        terminal_why = []
        for evaluator_name, function in self.terminal_when.items():
            if function(evaluation_dict[evaluator_name]):
                terminal = True
                terminal_why.append(evaluator_name)
        return terminal, terminal_why

    def _log_eval_average(self, result_dct_list):
        bresult = BenchmarkResult(result_dct_list, None)
        df = bresult.get_data_frame()
        grouped = df.apply(pd.to_numeric, errors='ignore').groupby(["scen_set", "behavior"]).mean()[
            self._evaluation_criteria()]
        self.logger.info("\n------------------- Current Evaluation Results ---------------------- \n Num. Results:{}\n {} \n \
---------------------------------------------------------------------".format(len(result_dct_list),
                                                                              grouped.to_string()))



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest

from bark.benchmark.benchmark_analyzer import BenchmarkAnalyzer
from bark.benchmark.benchmark_runner import BenchmarkResult

def dummy_benchmark_results():
    return [
      {"config_idx": 1, "collision":True, "metric1": 23.5, "behavior": "test", "scen_idx": 10},
      {"config_idx": 2, "collision":False, "metric1": 1, "behavior": "test2", "scen_idx": 12},
      {"config_idx": 3, "collision":False, "metric1": 23676.5, "behavior": "test1", "scen_idx": 3},
      {"config_idx": 121, "collision":False, "metric1": 1, "behavior": "test2", "scen_idx": 4},
      {"config_idx": 11, "collision":True, "metric1": 0.1, "behavior": "test", "scen_idx": 5},
      {"config_idx": 24, "collision":False, "metric1": 0.1, "behavior": "test1", "scen_idx": 7},
    ]

class BenchmarkAnalyzerTests(unittest.TestCase):
    def test_find_config_idx(self):
        brst = BenchmarkResult(result_dict=dummy_benchmark_results(), benchmark_configs=None)
        analyzer = BenchmarkAnalyzer(benchmark_result = brst)

        configs_found = analyzer.find_configs({"collision" : lambda x : x})
        self.assertEqual(configs_found, [1, 11])

        configs_found = analyzer.find_configs({"collision" : lambda x : not x})
        self.assertEqual(configs_found, [2, 3, 24, 121])

        configs_found = analyzer.find_configs({"metric1" : lambda x : x == 0.1})
        self.assertEqual(configs_found, [11, 24])

        configs_found = analyzer.find_configs({"collision" : lambda x : not x, "metric1" : lambda x : x == 0.1})
        self.assertEqual(configs_found, [24])

        configs_found = analyzer.find_configs({"collision" : lambda x : x, "metric1" : lambda x : x > 1})
        self.assertEqual(configs_found, [1])

        configs_found = analyzer.find_configs({"collision" : lambda x : x, "behavior" : lambda x : x=="test2"})
        self.assertEqual(configs_found, [])

        configs_found = analyzer.find_configs(scenario_idx_list=[12, 3 , 5])
        self.assertEqual(configs_found, [2, 3, 11])

        configs_found = analyzer.find_configs({"collision" : lambda x : not x}, scenario_idx_list=[12, 3 , 5])
        self.assertEqual(configs_found, [2, 3])

        configs_found = analyzer.find_configs({"collision" : lambda x : x}, scenarios_as_in_configs=[24, 11 , 121, 1])
        self.assertEqual(configs_found, [1, 11])


if __name__ == '__main__':
    unittest.main()


# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import time
import os
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.viewer.panda3d_easy import Panda3dViewer
from bark.runtime.viewer.video_renderer import VideoRenderer
from bark.runtime.scenario.scenario_generation.config_with_ease import \
  LaneCorridorConfig, ConfigWithEase
from bark.core.geometry.standard_shapes import CarLimousine
from bark.core.models.behavior import *
from bark.runtime.runtime import Runtime


# Parameters Definitions
param_server = ParameterServer()

param_server["BehaviorIDMLaneTracking"]["CrosstrackErrorGain"] = 2.5
param_server["BehaviorIDMClassic"]["DesiredVelocity"] = 5.
param_server["BehaviorIntersectionRuleBased"]["BrakingDistance"] = 10.
param_server["BehaviorIntersectionRuleBased"]["PredictionTimeHorizon"] = 5.
param_server["BehaviorIntersectionRuleBased"]["AngleDiffForIntersection"] = 0.25

# configure the lane corridors and how the agents 
lane_corridors = []
lane_corridors.append(
  LaneCorridorConfig(params=param_server,
                     source_pos=[-30, -3],
                     sink_pos=[30, -3],
                     behavior_model=BehaviorIntersectionRuleBased(param_server),
                     min_vel=5.,
                     max_vel=5.,
                     ds_min=5.,
                     ds_max=10.,
                     s_min=15.,
                     s_max=30.))
lane_corridors.append(
  LaneCorridorConfig(params=param_server,
                     source_pos=[30, 3],
                     sink_pos=[-30, 3],
                     behavior_model=BehaviorIntersectionRuleBased(param_server),
                     min_vel=5.,
                     max_vel=5.,
                     ds_min=5.,
                     ds_max=10.,
                     s_min=15.,
                     s_max=30.))
lane_corridors.append(
  LaneCorridorConfig(params=param_server,
                     source_pos=[3, -30],
                     sink_pos=[-30, 3],
                     behavior_model=BehaviorIntersectionRuleBased(param_server),
                     controlled_ids=True,
                     min_vel=5.,
                     max_vel=5.,
                     ds_min=5.,
                     ds_max=10.,
                     s_min=15.,
                     s_max=30.))

scenarios = \
  ConfigWithEase(num_scenarios=3,
                 map_file_name= os.path.join(os.path.dirname(__file__),"../runtime/tests/data/threeway_intersection.xodr"),
                 random_seed=0,
                 params=param_server,
                 lane_corridor_configs=lane_corridors)

# viewer
viewer = MPViewer(params=param_server, use_world_bounds=True)

# viewer = Panda3dViewer(params=param_server,
#                        x_range=[-40, 40],
#                        y_range=[-40, 40],
#                        follow_agent_id=agent3.id)

# World Simulation
sim_step_time = param_server["simulation"]["step_time",
                                           "Step-time used in simulation",
                                           0.2]
sim_real_time_factor = param_server["simulation"]["real_time_factor",
                                                  "execution in real-time or faster",
                                                  1.]
# viewer = VideoRenderer(renderer=viewer,
#                        world_step_time=sim_step_time,
#                        fig_path="/Users/hart/2019/bark/video")
env = Runtime(step_time=0.2,
              viewer=viewer,
              scenario_generator=scenarios,
              render=True)

# run 3 scenarios
for _ in range(0, 3):
  env.reset()
  # step each scenario 20 times
  for step in range(0, 80):
    env.step()
    time.sleep(sim_step_time/sim_real_time_factor)

# viewer.export_video(filename="/Users/hart/2019/bark/video/intersection", remove_image_dir=True)



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import json
import os
from bark.core.commons import Params
import logging
logging.getLogger().setLevel(logging.DEBUG)


class ParameterServer(Params):
    def __init__(self, **kwargs):
        Params.__init__(self)
        self.param_filename = None
        self.log_if_default = kwargs.pop("log_if_default", False)
        if "filename" in kwargs:
            self.load(kwargs["filename"])
            self.param_filename = kwargs["filename"]
        elif "json" in kwargs:
            self.ConvertToParam(kwargs["json"])
        else:
            self.store = dict()
        self.param_descriptions = dict()
        

    def __getitem__(self, key):
        if isinstance(key, tuple):
            new_key = key[0]
            self.param_descriptions[new_key] = key[1]
            if len(key) == 3:
                default_val = key[2]
            else:
                default_val = "MISSING"
        else:
            new_key = key
        if new_key in self.store:
            return self.store[new_key]
        else:
            if isinstance(key, tuple):  # if key for parameter
                self.store[new_key] = default_val
                if self.log_if_default:
                  logging.warning("Using default {} for {}".format(
                  default_val, new_key))
                return self.store[new_key]
            else:  # else it is a Params instance
                self.store[new_key] = ParameterServer(log_if_default = self.log_if_default)
                return self.store[new_key]

    def FindKey(self, param_key):
      delimiter = "::"
      for key, value in self.store.items():
                  if isinstance(value, ParameterServer):
                      found_key_tmp =  value.FindKey(param_key)
                      if found_key_tmp:
                        return "{}{}{}".format(key, delimiter,
                              found_key_tmp)
                  else:
                      if param_key == key:
                        return param_key
      return None


    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            new_key = key[0]
            self.param_descriptions[new_key] = key[1]
        else:
            new_key = key
        self.store[new_key] = value

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def load(self, fn):
        with open(fn) as file:
            dict = json.load(file)
        self.ConvertToParam(dict)

    def ConvertToParam(self, new_dict):
        self.store = dict()
        for key, value in new_dict.items():
            if isinstance(value, dict):
                param = ParameterServer(log_if_default = self.log_if_default)
                self.store[key] = param.ConvertToParam(value)
            else:
                self.GetValFromString(key, "", value, log_if_default=False)

        return self

    def ConvertToDict(self, print_description=False):
        dict = {}
        for key, value in self.store.items():
            if isinstance(value, ParameterServer):
                v = value.ConvertToDict(print_description)
                if len(v) == 0:
                    if print_description:
                        if key in self.param_descriptions:
                            dict[key] = self.param_descriptions[key]
                        else:
                            dict[key] = "--"
                    else:
                        dict[key] = "MISSING"
                else:
                    dict[key] = v
            else:
                dict[key] = value

                if print_description:
                    if key in self.param_descriptions:
                        dict[key] = self.param_descriptions[key]
                    else:
                        dict[key] = "--"

        return dict

    def Save(self, filename, print_description=False):
        #if not os.path.exists(os.path.dirname(filename)):
            #try:
           #     os.makedirs(os.path.dirname(filename))
          #  except:
         #       print("Could not dump parameters, no rights to create file directory.")
        #        return
        with open(filename, 'w') as outfile:
            print("Writing parameters to {}".format(os.path.abspath(filename)))
            outfile.write(
                json.dumps(
                    self.ConvertToDict(print_description=print_description),
                    indent=4))

    def get_val_iter(self, hierarchy, description, default_value):
        if not hierarchy:
            raise ValueError("PARAM HIERARCHY IS EMPTY")
        else:
            key = hierarchy[0]
            hierarchy.pop(0)
            if key in self.store:
                #if isinstance(self.store[key], ParameterServer):
                if type(self.store[key]) == ParameterServer:
                    #raise ValueError("get here")
                    return self.store[key].get_val_iter(
                        hierarchy, description, default_value)
                else:
                    return self.store[key], False
            else:
                if not hierarchy:
                    self.store[key] = default_value
                    return default_value, True
                else:
                    self.store[key] = ParameterServer(log_if_default = self.log_if_default)
                    return self.store[key].get_val_iter(
                        hierarchy, description, default_value)
        return

    def GetCondensedParamList(self):
        def check_append(value):
          if isinstance(value, float) or isinstance(value, int) \
              or isinstance(value,bool):
              return True
          # list float
          elif isinstance(value, list) and all(isinstance(el, float) for el in value):
            return True
          # list list float
          elif isinstance(value, list):
              for el in value:
                if not isinstance(el, list):
                  return False
                for e in el:
                  if not isinstance(e, float):
                    return False
              return True
          return False
        hierarchy_delimiter = "::"
        condensed_param_list = []
        for key, value in self.store.items():
            if isinstance(value, ParameterServer):
                child_list = value.GetCondensedParamList()
                for param_tuple in child_list:
                    param_name = "{}{}{}".format(key,
                                    hierarchy_delimiter, param_tuple[0])
                    param_value = param_tuple[1]
                    if check_append(param_value):
                        condensed_param_list.append((param_name, param_value))
            else:
                if check_append(value):
                    condensed_param_list.append((key, value))
        test = condensed_param_list
        return condensed_param_list

    def GetValFromString(self, hierarchy, description, default_value, log_if_default):
        hierarchy = [x.strip() for x in hierarchy.split("::")]
        if log_if_default:
          # first search for key, otherwise default already integrated into store
          found_key = self.FindKey(hierarchy[-1])
        value, used_default = self.get_val_iter(hierarchy.copy(), description, default_value)
        if log_if_default and used_default:
          logging.warning("Using default {} for {}".format(
                                default_value, hierarchy))
          if found_key:
            logging.warning("Did you mean param {}".format(found_key))
          else:
            logging.warning("Key was nowhere found.")
        return value
        
    # get values
    def GetBool(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetReal(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetInt(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetListListFloat(self, param_name, description, default_value):
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)
    
    def GetListFloat(self, param_name, description, default_value):
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def access(self, param_name):
        return self[param_name]

    def SetBool(self, param_name, value):
        self[param_name] = value
        return

    def SetReal(self, param_name, value):
        self[param_name] = value
        return

    def SetInt(self, param_name, value):
        self[param_name] = value
        return

    def AddChild(self, name):
        return self.__getitem__(name)



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.runtime.scenario import Scenario
from bark.runtime.scenario.scenario_generation import ScenarioGeneration
from bark.runtime.scenario.interaction_dataset_processing.interaction_dataset_reader import agent_from_trackfile
from bark.runtime.scenario.interaction_dataset_processing import DatasetDecomposer
from bark.runtime.commons import ParameterServer
# PyBind imports
from bark.core.world.map import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *


class InteractionDatasetScenarioGenerationFull(ScenarioGeneration):
  # This class reads in a track file from the interaction dataset
  # and generates a scenario for each agent as the eval agent.

    def __init__(self, params=None, num_scenarios=None, random_seed=None):
        super().__init__(params, num_scenarios, random_seed)
        self.initialize_params(params)

    def initialize_params(self, params):
        super().initialize_params(params)
        params_temp = \
            self._params["Scenario"]["Generation"]["InteractionDatasetScenarioGenerationFull"]
        self._map_file_name = params_temp["MapFilename",
                                          "Path to the open drive map",
                                          "bark/runtime/tests/data/DR_DEU_Merging_MT_v01_shifted.xodr"]
        self._track_file_name = params_temp["TrackFilename",
                                            "Path to track file (csv)",
                                            "bark/runtime/tests/data/interaction_dataset_dummy_track.csv"]
        self._behavior_models = params_temp["BehaviorModel",
                                            "Overwrite static trajectory with prediction model", {}]

    # TODO: remove code duplication with configurable scenario generation
    def create_scenarios(self, params, num_scenarios):
        """
            see baseclass
        """
        scenario_list = []

        dataset_decomposer = DatasetDecomposer(map_filename=self._map_file_name,
                                               track_filename=self._track_file_name)
        scenario_track_info_list = dataset_decomposer.decompose()

        # for scenario_idx in range(0, num_scenarios):
        for idx_s, scenario_track_info in enumerate(scenario_track_info_list):
            if idx_s < num_scenarios:
                try:
                    scenario = self.__create_single_scenario__(
                        scenario_track_info)
                except:
                    raise ValueError(
                        "Generation of scenario failed: {}".format(scenario_track_info))
                scenario_list.append(scenario)
            else:
                break
        return scenario_list

    def __create_single_scenario__(self, scenario_track_info):
        scenario_track_info.TimeSanityCheck()
        
        scenario = Scenario(map_file_name=self._map_file_name,
                            json_params=self._params.ConvertToDict())
        world = scenario.GetWorldState()
        agent_list = []
        track_params = ParameterServer()
        track_params["execution_model"] = 'ExecutionModelInterpolate'
        track_params["dynamic_model"] = 'SingleTrackModel'
        track_params["map_interface"] = world.map

        for id_other in scenario_track_info.GetOtherTrackInfos().keys():
            if str(id_other) in self._behavior_models:
                track_params["behavior_model"] = self._behavior_models[str(
                    id_other)]
            else:
                track_params["behavior_model"] = None
            agent = agent_from_trackfile(
                track_params, self._params, scenario_track_info, id_other)
            agent_list.append(agent)

        id_ego = scenario_track_info.GetEgoTrackInfo().GetTrackId()
        if str(id_ego) in self._behavior_models:
            track_params["behavior_model"] = self._behavior_models[str(id_ego)]
        else:
            track_params["behavior_model"] = None
        agent = agent_from_trackfile(
            track_params, self._params, scenario_track_info, id_ego)
        agent_list.append(agent)

        scenario._agent_list = agent_list  # must contain all agents!
        scenario._eval_agent_ids = [
            scenario_track_info.GetEgoTrackInfo().GetTrackId()]

        return scenario


# Copyright (c) 2020 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo


class ScenarioTrackInfo:
    def __init__(self, map_filename, track_filename, ego_track_info, start_ts=None, end_ts=None):
        self._map_filename = map_filename
        self._track_filename = track_filename
        self._ego_track_info = ego_track_info

        if start_ts is None:
            self._start_ts = ego_track_info.GetStartOffset()
        else:
            self._start_ts = start_ts

        if end_ts is None:
            self._end_ts = ego_track_info.GetEndOffset()
        else:
            self._end_ts = end_ts

        self._other_agents_track_infos = {}

    def __str__(self):
        str_out = 'start_ts={} end_ts={} ego: {}'.format(self.GetStartTs(), self.GetEndTs(), self._ego_track_info)
        for other in self.GetOtherTrackInfos().values():
            str_out = str_out + ', other: {}'.format(other)
        return str_out

    def AddTrackInfoOtherAgent(self, track_info_other):
        self._other_agents_track_infos[track_info_other.GetTrackId(
        )] = track_info_other

    def GetMapFilename(self):
        return self._map_filename

    def GetTrackFilename(self):
        return self._track_filename

    def GetEgoTrackInfo(self):
        return self._ego_track_info

    def GetStartTs(self):
        return self._start_ts

    def GetEndTs(self):
        return self._end_ts

    def GetOtherTrackInfos(self):
        return self._other_agents_track_infos

    def TimeSanityCheck(self):
        # for other in self.GetOtherTrackInfos().values():
        #     if self.GetStartTs() > other.GetStartOffset():
        #         raise ValueError(
        #             "Other agent {} starts before scenario".format(other.GetTrackId()))
        #     elif self.GetEndTs() < other.GetEndOffset():
        #         raise ValueError(
        #             "Other agent {} ends after scenario".format(other.GetTrackId()))

        if self.GetStartTs() < self.GetEgoTrackInfo().GetStartOffset():
            raise ValueError("ego agent {} starts after scenario".format(
                self.GetEgoTrackInfo().GetTrackId()))
        elif self.GetEndTs() > self.GetEgoTrackInfo().GetEndOffset():
            raise ValueError("ego agent {} ends before scenario".format(
                self.GetEgoTrackInfo().GetTrackId()))

        return True



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
import os
import ray

from load.benchmark_database import BenchmarkDatabase
from serialization.database_serializer import DatabaseSerializer
from bark.benchmark.benchmark_runner import BenchmarkRunner, BenchmarkConfig
from bark.benchmark.benchmark_runner_mp import BenchmarkRunnerMP, _BenchmarkRunnerActor, \
        deserialize_benchmark_config, serialize_benchmark_config

from bark.runtime.scenario.scenario_generation.configurable_scenario_generation \
  import ConfigurableScenarioGeneration

from bark.runtime.viewer.matplotlib_viewer import MPViewer

from bark.core.world.evaluation import *
from bark.runtime.commons.parameters import ParameterServer
from bark.core.models.behavior import BehaviorIDMClassic, BehaviorConstantVelocity

# to find database files
os.chdir("../benchmark_database/")

class DatabaseRunnerTests(unittest.TestCase):
    #@unittest.skip
    def test_database_runner(self):
      dbs = DatabaseSerializer(test_scenarios=2, test_world_steps=2, num_serialize_scenarios=5)
      cwd = os.getcwd()
      dbs.process("data/database1")
      local_release_filename = dbs.release(version="test")

      db = BenchmarkDatabase(database_root=local_release_filename)
      evaluators = {"success" : "EvaluatorGoalReached", "collision" : "EvaluatorCollisionEgoAgent",
                    "max_steps": "EvaluatorStepCount"}
      terminal_when = {"collision" :lambda x: x, "max_steps": lambda x : x>5}
      params = ParameterServer() # only for evaluated agents not passed to scenario!
      behaviors_tested = {"IDM": BehaviorIDMClassic(params), "Const" : BehaviorConstantVelocity(params)}

      benchmark_runner = BenchmarkRunner(benchmark_database=db,
                                         evaluators=evaluators,
                                         terminal_when=terminal_when,
                                         behaviors=behaviors_tested,
                                         log_eval_avg_every=1)

      result = benchmark_runner.run()
      df = result.get_data_frame()
      print(df)
      self.assertEqual(len(df.index), 10) # 2 Behaviors * 2 Serialize Scenarios * 5 scenario sets

    @unittest.skip
    def test_database_multiprocessing_runner(self):
      dbs = DatabaseSerializer(test_scenarios=4, test_world_steps=5, num_serialize_scenarios=10)
      dbs.process("data/database1")
      local_release_filename = dbs.release(version="test")

      db = BenchmarkDatabase(database_root=local_release_filename)
      evaluators = {"success" : "EvaluatorGoalReached", "collision" : "EvaluatorCollisionEgoAgent",
                    "max_steps": "EvaluatorStepCount"}
      terminal_when = {"collision" :lambda x: x, "max_steps": lambda x : x>5}
      params = ParameterServer() # only for evaluated agents not passed to scenario!
      behaviors_tested = {"IDM": BehaviorIDMClassic(params), "Const" : BehaviorConstantVelocity(params)}

      benchmark_runner = BenchmarkRunnerMP(benchmark_database=db,
                                          evaluators=evaluators,
                                          terminal_when=terminal_when,
                                          behaviors=behaviors_tested,
                                          log_eval_avg_every=10)

      result = benchmark_runner.run()

      params2 = ParameterServer()
      viewer = MPViewer(
            params=params2,
            x_range=[5060, 5160],
            y_range=[5070,5150],
            use_world_bounds=True)
      rst, _ = benchmark_runner.run_benchmark_config(10, viewer=viewer)


      rst = benchmark_runner.run(maintain_history=True)
      self.assertEqual(len(rst.get_histories()), 40)

      rst, scenario_history = benchmark_runner.run_benchmark_config(11, viewer=None, maintain_history=True)
      print(scenario_history)
      viewer = MPViewer(
            params=params2,
            x_range=[5060, 5160],
            y_range=[5070,5150],
            use_world_bounds=True)
      viewer.drawWorld(world=scenario_history[5].GetWorldState(),
                        eval_agent_ids=scenario_history[5].eval_agent_ids)

      viewer.show(block=True)


      df = result.get_data_frame()
      print(df)
      self.assertEqual(len(df.index), 40) # 2 Behaviors * 10 Serialize Scenarios * 2 scenario sets

if __name__ == '__main__':
    unittest.main()





# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.


import numpy as np

from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderControlledAgents

from bark.core.world.goal_definition import GoalDefinition, GoalDefinitionPolygon, GoalDefinitionStateLimits
from bark.runtime.commons.parameters import ParameterServer

# no one (in this road corridor) is a controlled agent
class NoneControlled(ConfigReaderControlledAgents):
   # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    return [False] * len(agent_states), {}, config_param_object

# Select one agent randomly as controlled agent
class RandomSingleAgent(ConfigReaderControlledAgents):
  # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    controlled_agent_idx = self.random_state.randint(low=0, high=len(agent_states), size=None) # todo(@bernhard): make seed state global, a.t.m. always same number drawn
    controlled_list = [False] * len(agent_states)
    controlled_list[controlled_agent_idx] = True
    return controlled_list, {}, config_param_object


from setuptools import setup, find_packages, Extension
import sys

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = "bark-simulator",
    version = "0.0.92",
    description = "A tool for Behavior benchmARKing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers = ["Development Status :: 1 - Planning", "Intended Audience :: Developers", "License :: OSI Approved :: MIT License", "Operating System :: POSIX", "Programming Language :: Python :: 3.7"],
    keywords = "simulator autonomous driving machine learning",
    url = "https://github.com/bark-simulator/bark",
    author = "Fortiss",
    author_email = "cheema@fortiss.org",
    license = "MIT",
    packages=find_packages(),
    install_requires=[
    'matplotlib>=3.0.3',
    'numpy>=1.18.1',
    'lxml>=4.4.2',
    'scipy>=1.4.1',
    'sphinx>=2.3.1',
    'recommonmark>=0.6.0',
    'sphinx_rtd_theme>=0.4.3',
    'pandas>=0.24.2',
    'autopep8>=1.4.4',
    'cpplint>=1.4.4',
    'pygame>=1.9.6',
    'panda3d>=1.10.5',
    'aabbtree>=2.3.1',
    'ray>=0.7.6',
    'psutil>=5.6.6',
    'notebook>=6.0.3',
    'jupyter>=1.0.0',
    'ipython>=7.13.0'
    ],
    ext_modules=[
        Extension(
            name='',
            sources=[]
        )
    ] if sys.platform != 'linux' else [],
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.7',
)




# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import os
import shutil

from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.scenario.interaction_dataset_processing.interaction_dataset_reader import agent_from_trackfile, trajectory_from_track
from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo
from bark.runtime.scenario.interaction_dataset_processing.scenario_track_info import ScenarioTrackInfo
from bark.runtime.scenario.scenario import Scenario
from bark.core.geometry import Point2d, Collide

from com_github_interaction_dataset_interaction_dataset.python.utils import dataset_reader



class DatasetDecomposer:
    def __init__(self, map_filename, track_filename):
        self._map_filename = map_filename
        self._track_filename = track_filename
        self._track_dict = dataset_reader.read_tracks(track_filename)
        self._map_interface = self.__setup_map_interface__()
        self._agents_track_infos = self.__setup_agents_track_infos__()

    def __setup_map_interface__(self):
        params = ParameterServer()
        # we are creating a dummy scenario to get the map interface from it
        scenario = Scenario(map_file_name=self._map_filename,
                            json_params=params.ConvertToDict())
        world = scenario.GetWorldState()
        map_interface = world.map
        return map_interface

    def __find_first_ts_on_map__(self, id_ego):
        traj = trajectory_from_track(self._track_dict[id_ego])
        for state in traj:
            point_agent = Point2d(state[1], state[2])
            lane_list = self._map_interface.find_nearest_lanes(point_agent, 3)
            for lane in lane_list:
                polygon = self._map_interface.GetRoadgraph().GetLanePolygonForLaneId(lane.lane_id)
                if Collide(polygon, point_agent):
                    time_ego_first = state[0]*1e3  # use timestamp in ms
                    return time_ego_first

        return None

    def __setup_agents_track_infos__(self):
        # dictionary mapping first valid timestamp to agnet id
        agents_track_infos = {}
        for agent_id in self._track_dict.keys():
            # TODO: this could be made optional
            first_ts_on_map = self.__find_first_ts_on_map__(agent_id)
            if first_ts_on_map is None:
                print("Agent %d not found on map" % agent_id)
                pass
            else:
                start_offset = first_ts_on_map
                end_offset = self._track_dict[agent_id].time_stamp_ms_last
                new_agent = AgentTrackInfo(filename=self._track_filename, track_id=agent_id,
                                           start_offset=start_offset, end_offset=end_offset)
                agents_track_infos[agent_id] = new_agent
        return agents_track_infos

    def __get_agent_track_info__(self, agent_id):
        if agent_id in self._agents_track_infos.keys():
            return self._agents_track_infos[agent_id]
        else:
            raise ValueError("agent {} not available".format(agent_id))

    def __find_all_ids__(self, id_ego):
        # for each agent extract ids of other agents present in the same time span
        list_ids = []
        time_ego_first = self.__get_agent_track_info__(id_ego).GetStartOffset()
        time_ego_last = self.__get_agent_track_info__(id_ego).GetEndOffset()

        for id_current in self._agents_track_infos.keys():
            if id_ego == id_current:
                pass
            elif self.__get_agent_track_info__(id_current).GetEndOffset() < time_ego_first:
                # other ends too early
                pass
            elif self.__get_agent_track_info__(id_current).GetStartOffset() > time_ego_last:
                # other starts too late
                pass
            else:
                list_ids.append(id_current)

        return list_ids

    def __find_all_scenarios__(self):
        scenario_list = []
        for id_ego in self._agents_track_infos.keys():
            ego_track_info = self.__get_agent_track_info__(id_ego)
            new_scenario = ScenarioTrackInfo(
                map_filename=self._map_filename, track_filename=self._track_filename, ego_track_info=ego_track_info)

            ids_others = self.__find_all_ids__(id_ego)
            for id_o in ids_others:
                agent_o_track_info = self.__get_agent_track_info__(id_o)
                new_scenario.AddTrackInfoOtherAgent(agent_o_track_info)

            scenario_list.append(new_scenario)

        return scenario_list

    def decompose(self):

        scenario_list = self.__find_all_scenarios__()
        return scenario_list



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from maude_interface import make_initial_world, get_ego_agent, apply_action_to_ego_agent
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.video_renderer import VideoRenderer
import time

# list of tuples with acceleration / steering angle
primitives = [(0, 0), (-5, 0), (5, 0)]
world = make_initial_world(primitives)

ego_agent = get_ego_agent(world)
apply_action_to_ego_agent(world, 0) # applies action with idx 0 (keep velocity)

params = ParameterServer()
viewer = MPViewer(params=params, use_world_bounds=True)

# World Simulation
sim_step_time = params["simulation"]["step_time",
                                     "Step-time in simulation", 0.2]
sim_real_time_factor = params["simulation"]["real_time_factor",
                                            "execution in real-time or faster", 1]

# Draw map
video_renderer = VideoRenderer(renderer=viewer, world_step_time=sim_step_time)

for _ in range(0, 20):
    world.Step(sim_step_time)
    viewer.clear()
    viewer.drawWorld(world)
    viewer.drawGoalDefinition(ego_agent.goal_definition, "red", 0.5, "red")

    print(ego_agent.state)

    # applies action with idx 1 (break)
    apply_action_to_ego_agent(world, 1)

    time.sleep(sim_step_time/sim_real_time_factor)



#!/usr/bin/env python

import matplotlib
import matplotlib.axes
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection


def set_visible_area(laneletmap, axes):
    min_x = 10e9
    min_y = 10e9
    max_x = -10e9
    max_y = -10e9

    for point in laneletmap.pointLayer:
        min_x = min(point.x, min_x)
        min_y = min(point.y, min_y)
        max_x = max(point.x, max_x)
        max_y = max(point.y, max_y)

    axes.set_aspect('equal', adjustable='box')
    axes.set_xlim([min_x - 10, max_x + 10])
    axes.set_ylim([min_y - 10, max_y + 10])


def draw_lanelet_map(laneletmap, axes):

    assert isinstance(axes, matplotlib.axes.Axes)

    set_visible_area(laneletmap, axes)

    unknown_linestring_types = list()

    for ls in laneletmap.lineStringLayer:

        if "type" not in ls.attributes.keys():
            raise RuntimeError("ID " + str(ls.id) + ": Linestring type must be specified")
        elif ls.attributes["type"] == "curbstone":
            type_dict = dict(color="black", linewidth=1, zorder=10)
        elif ls.attributes["type"] == "line_thin":
            if "subtype" in ls.attributes.keys() and ls.attributes["subtype"] == "dashed":
                type_dict = dict(color="white", linewidth=1, zorder=10, dashes=[10, 10])
            else:
                type_dict = dict(color="white", linewidth=1, zorder=10)
        elif ls.attributes["type"] == "line_thick":
            if "subtype" in ls.attributes.keys() and ls.attributes["subtype"] == "dashed":
                type_dict = dict(color="white", linewidth=2, zorder=10, dashes=[10, 10])
            else:
                type_dict = dict(color="white", linewidth=2, zorder=10)
        elif ls.attributes["type"] == "pedestrian_marking":
            type_dict = dict(color="white", linewidth=1, zorder=10, dashes=[5, 10])
        elif ls.attributes["type"] == "bike_marking":
            type_dict = dict(color="white", linewidth=1, zorder=10, dashes=[5, 10])
        elif ls.attributes["type"] == "stop_line":
            type_dict = dict(color="white", linewidth=3, zorder=10)
        elif ls.attributes["type"] == "virtual":
            type_dict = dict(color="blue", linewidth=1, zorder=10, dashes=[2, 5])
        elif ls.attributes["type"] == "road_border":
            type_dict = dict(color="black", linewidth=1, zorder=10)
        elif ls.attributes["type"] == "guard_rail":
            type_dict = dict(color="black", linewidth=1, zorder=10)
        elif ls.attributes["type"] == "traffic_sign":
            continue
        else:
            if ls.attributes["type"] not in unknown_linestring_types:
                unknown_linestring_types.append(ls.attributes["type"])
            continue

        ls_points_x = [pt.x for pt in ls]
        ls_points_y = [pt.y for pt in ls]

        plt.plot(ls_points_x, ls_points_y, **type_dict)

    if len(unknown_linestring_types) != 0:
        print("Found the following unknown types, did not plot them: " + str(unknown_linestring_types))

    lanelets = []
    for ll in laneletmap.laneletLayer:
        points = [[pt.x, pt.y] for pt in ll.polygon2d()]
        polygon = Polygon(points, True)
        lanelets.append(polygon)

    ll_patches = PatchCollection(lanelets, facecolors="lightgray", edgecolors="None", zorder=5)
    axes.add_collection(ll_patches)

    if len(laneletmap.laneletLayer) == 0:
        axes.patch.set_facecolor('lightgrey')

    areas = []
    for area in laneletmap.areaLayer:
        if area.attributes["subtype"] == "keepout":
            points = [[pt.x, pt.y] for pt in area.outerBoundPolygon()]
            polygon = Polygon(points, True)
            areas.append(polygon)

    area_patches = PatchCollection(areas, facecolors="darkgray", edgecolors="None", zorder=5)
    axes.add_collection(area_patches)



import numpy as np
from direct.showbase.ShowBase import ShowBase
from direct.gui.DirectGui import DirectButton
from direct.task import Task
from math import pi
import os

from bark.core.viewer import Viewer
from bark.core.geometry import Line2d, Point2d
from bark.core.models.dynamic import StateDefinition
from bark.runtime.viewer.viewer import BaseViewer
from panda3d.core import *

class Panda3dViewer(BaseViewer, ShowBase):
  def __init__(self, params=None, **kwargs):
    super(Panda3dViewer, self).__init__(params=params, **kwargs)
    self._model_path = kwargs.pop(
      "model_path",
      os.path.join(os.path.dirname(os.path.abspath(__file__))) + "/models/car_model.obj")
    self._model_scale = kwargs.pop("model_scale",
                                   np.array([.3,.3,.3], dtype=float))
    self._model_orientation = \
      kwargs.pop("model_orientation",
                 np.array([0, 90, 90],
                 dtype=float))
    self._model_translation = kwargs.pop("model_translation",
                                         np.array([0., 0., 0.], dtype=float))
    self._screen_dims = kwargs.pop("screen_dims", [2*1024, 1024])
    self._line_thickness = kwargs.pop("line_thickness", 0.05)
    self._follow_agent = kwargs.pop("follow_agent", True)
    self._follow_agent_id = kwargs.pop("follow_agent_id", None)
    self._camera_pose = kwargs.pop("camera_pose", [0, 20, 60])
    self._light_pose = kwargs.pop("light_pose", [0, 0, 10000])
    self._camera_orientation = kwargs.pop("camera_pose", [0, 270, 0])
    self._frame_count = 0

    # Panda3d
    self._window = WindowProperties().getDefault()
    self._window.setSize(self._screen_dims[0], self._screen_dims[1])
    self._window.setTitle("BARK Viewer")
    WindowProperties.setDefault(self._window)
    ShowBase.__init__(self)

    self.SetupBasics()
    self._agent_nodes = {}
  
  def SetupBasics(self):
    self.AddCamera()
    self.AddPlane()
    self.AddLight()
    self.AddGenerator()

  def AddCamera(self):
    base.disableMouse()
    # TODO: calculate z based on viewing range

    # self.camera is available due to ShowBase class
    self.camera.setPos(
      self._camera_pose[0],
      self._camera_pose[1],
      self._camera_pose[2])
    self.camera.setHpr(
      self._camera_orientation[2],
      self._camera_orientation[1],
      self._camera_orientation[0])
    
  def AddPlane(self, color=VBase4(1, 1, 1, 1)):
    card = CardMaker("plane")
    card.set_color(color)
    card.setFrame(-10000, 10000, -10000, 10000)
    n = NodePath()
    self.plane = n.attach_new_node(card.generate())
    self.plane.setHpr(0, 270, 0)
    self.plane.reparentTo(self.render)

  def AddLight(self):
    alight = AmbientLight('alight')
    alight.setColor(VBase4(.1, .1, .1, 1))
    self._ambient_light = self.render.attachNewNode(alight)
    self.render.setLight(self._ambient_light)

    # TODO(@hart): position of pt-light needs to be fixed
    plight = PointLight('plight')
    plight.setColor((1., 1., 1., 1))
    self._point_light = self.render.attachNewNode(plight)
    self._point_light.setPos(self._light_pose[0], self._light_pose[1], self._light_pose[2])
    self.render.setLight(self._point_light)

  def AddGenerator(self, budget=1000):
    self._generator = MeshDrawer()
    self._generator.setBudget(budget)
    self._generator_node = self._generator.getRoot()
    self._generator_node.reparentTo(self.render)

  def drawPolygon2d(self, polygon, color, alpha, facecolor=None):
    points = polygon.ToArray()
    self.drawLine2d(polygon, color, alpha)
  
  def drawLine2d(self, line2d, color='blue', alpha=1.0, dashed=False, zorder=1, linewidth=1):
    # TODO: calculate the line thickness based on the z-value
    line2d_np = line2d.ToArray()
    for point in line2d_np:
      self._generator.link_segment(
        Vec3(point[0], point[1], 0.1),
        Vec4(0, 0, 1, 1),
        self._line_thickness,
        self.getColor(Viewer.Color.Blue))
    self._generator.link_segment_end(Vec4(0, 0, 1, 1), self.getColor(Viewer.Color.Blue))

  def get_aspect_ratio(self):
    return self._screen_dims[0] / self._screen_dims[1]
  
  def drawAgent(self, agent, color, alpha, facecolor):
    x_pos = agent.state[int(StateDefinition.X_POSITION)]
    y_pos = agent.state[int(StateDefinition.Y_POSITION)]
    if not agent.id in self._agent_nodes:
      self._agent_nodes[agent.id] = self.loader.loadModel(
        self._model_path)
      self._agent_nodes[agent.id].reparentTo(self.render)
      self._agent_nodes[agent.id].setScale(
        self._model_scale[0], self._model_scale[1], self._model_scale[2])
      self._agent_nodes[agent.id].setColor(self.getColor(Viewer.Color.Blue))
      self._agent_nodes[agent.id].setColorScale(self.getColor(Viewer.Color.Blue))

    angle = ((agent.state[int(StateDefinition.THETA_POSITION)] * 180) / pi +
              self._model_orientation[0])
    self._agent_nodes[agent.id].setHpr(
      angle,
      self._model_orientation[1],
      self._model_orientation[2])
    self._agent_nodes[agent.id].setPos(
      x_pos + self._model_translation[0],
      y_pos + self._model_translation[1],
      self._model_translation[2])

  def getColor(self, color):
    if isinstance(color, Viewer.Color):
      return {
        Viewer.Color.White:
        Vec4(1, 1, 1, 1),
        Viewer.Color.Red:
        Vec4(1, 0, 0, 1),
        Viewer.Color.Blue:
        Vec4(0, 0, 1, 1),
        Viewer.Color.Magenta:
        Vec4(100.0 / 255.0, 1, 100.0 / 255.0, 1),
        Viewer.Color.Brown:
        Vec4(150.0 / 255.0, 50.0 / 255.0, 50.0 / 255.0, 1),
        Viewer.Color.Black:
        Vec4(0, 0, 0, 1)
      }.get(color, Vec4(0, 0, 0, 1))
    else:
      return Vec4(color[0], color[1], color[2], 1)

  def drawWorld(self, world, eval_agent_ids=None, filename=None, scenario_idx=None, debug_text=True):
    self._generator.begin(base.cam, self.render)
    super(Panda3dViewer, self).drawWorld(world, eval_agent_ids, filename, scenario_idx, debug_text)
    
    if self._follow_agent_id is not None:
      self.camera.lookAt(self._agent_nodes[self._follow_agent_id])
    elif isinstance(self._follow_agent, bool) and self._follow_agent == True:
      self.camera.lookAt(self._agent_nodes[eval_agent_ids[0]])

    # self.camera.setHpr(
    #   self._camera_orientation[2],
    #   self._camera_orientation[1],
    #   self._camera_orientation[0])
    base.win.saveScreenshot("/Users/hart/2019/bark/screenshots/screen_{:02d}.png".format((self._frame_count)))
    self._frame_count += 1
    self._generator.end()
    self.taskMgr.step()

  def Reset(self):
    for agent_node in self._agent_nodes.values():
      agent_node.removeNode()
    self._agent_nodes = {}


# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
from bark.core.commons import *
from bark.runtime.commons.parameters import ParameterServer


class ParamsTests(unittest.TestCase):
    def test_parameters(self):
        # initialize Params
        p = ParameterServer()  # TODO: this has to be the ParameterServer

        # set new parameter
        self.assertTrue(p["LetsTest"]["hierarchy", "bla", True])

        # check whether right value is recovered
        tester = p["Car"]["Length", "Car Length", 6]
        tester2 = p["Car"]["Length", "Car Length", 8]
        self.assertEqual(tester, 6)
        self.assertEqual(
            tester2,
            6)  # value should not change, since key already exists in dict

        # check whether access without description and default value is possible
        self.assertEqual(p["Car"]["Length"], 6)

        # check whether setting values works
        p["Age"] = 24
        self.assertEqual(p["Age"], 24)
        p["Localization"]["Number of Particles"] = 2000
        self.assertEqual(p["Localization"]["Number of Particles"], 2000)

        # C++ Test in /bark/commons/Params/params_test.h
        # write in parameters in C++ and check whether they can be accessed in python afterwards
        #ParamsTest(p)
        #self.assertEqual(p["param_cpp"], 16.5)

        # add child in python
        child = p.AddChild("ch")
        self.assertTrue(child["ChildTest"]["hierarchy", "bla", True])

        # write parameters to json file
        p.Save("written_a_param_test.json")


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.



from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderBehaviorModels

from bark.core.models.behavior import *
from bark.runtime.commons.parameters import ParameterServer

  # this config reader defines behavior models with fixed type for all agents
class FixedBehaviorType(ConfigReaderBehaviorModels):
  def __init__(self, random_state):
    super().__init__(random_state)
    self.param_servers = []

  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    model_type = config_param_object["ModelType", "Type of behavior model \
                used for all vehicles", "BehaviorIDMClassic"]
    model_params = config_param_object.AddChild("ModelParams")
    self.param_servers.append(model_params) # use the same param server for all models
    behavior_models = []
    behavior_model_types = []
    model_default_params = ParameterServer(json={})
    for _ in agent_states:
      bark_model, params = self.model_from_model_type(model_type, model_params)
      behavior_models.append(bark_model)
      behavior_model_types.append(model_type)
    return behavior_models, {"behavior_model_types" : behavior_model_types}, config_param_object

  def model_from_model_type(self, model_type, params):
    bark_model = eval("{}(params)".format(model_type))    
    return bark_model, params

  def get_param_servers(self):
    return self.param_servers


class SampleBehaviorType(ConfigReaderBehaviorModels):
  def __init__(self, random_state):
    super().__init__(random_state)
    self.param_servers = []

  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    model_types = config_param_object["ModelTypesList", "Type of behavior model" \
                "used for all vehicles", ["BehaviorIDMClassic", "BehaviorMobil"]]
    model_params = config_param_object.AddChild("ModelParams")
    # ----- DEFAULT PARAMETER HANDLING
    # based on types retrieve default params which are maintained as scenario defaults
    for model_type in model_types:
        behavior_params = model_params.AddChild(model_type)
        _, _ = self.model_from_model_type(model_type, behavior_params)
        #param server must be persisted for each behavior to enable serialization of parameters
        

    #------ BEHAVIOR MODEL SAMPLING
    behavior_models = []
    behavior_model_types = []
    for _ in agent_states:
        model_idx = self.random_state.randint(low=0, high=len(model_types), size=None) 
        model_type = model_types[model_idx]
        model_type_params = model_params.AddChild(model_type)
        params = ParameterServer()
        bark_model, params = self.model_from_model_type(model_type, model_type_params)
        self.param_servers.append(model_type_params) 
        behavior_models.append(bark_model)
        behavior_model_types.append(model_type)
    return behavior_models, {"behavior_model_types" : behavior_model_types}, config_param_object

  def model_from_model_type(self, model_type, params):
    bark_model = eval("{}(params)".format(model_type))    
    return bark_model, params

  def get_param_servers(self):
    return self.param_servers


import numpy as np
from direct.showbase.ShowBase import ShowBase
from direct.gui.DirectGui import DirectButton
from direct.task import Task
from math import pi
import os

from bark.core.viewer import Viewer
from bark.core.models.dynamic import StateDefinition
from bark.runtime.viewer.viewer import BaseViewer
from panda3d.core import WindowProperties, MeshDrawer, VBase4, VBase3,Vec4, Vec3, Mat4, AmbientLight, CardMaker, NodePath


class Panda3dViewer(BaseViewer, ShowBase):
    def __init__(self, params=None, **kwargs):
        # Load 3d-Model and parameter
        super(Panda3dViewer, self).__init__(params=params)
        self.world_x_range = kwargs.pop("x_range", [-40, 40])
        self.world_y_range = kwargs.pop("y_range", [-40, 40])
        self.follow_agent_id = kwargs.pop("follow_agent_id", True)
        self.perspective = kwargs.pop("perspective", -3)
        self.screen_dims = kwargs.pop("screen_dims", [1024, 1024])
        self.path = os.path.join(os.path.dirname(os.path.abspath(__file__)))
        self.agent_model_path = kwargs.pop("model_path",
                                           self.path + "/models/car.obj")
        self.texture_path = kwargs.pop("model_path",
                                           self.path + "/models/white_texture.png")
        self.agent_scale = kwargs.pop("model_scale",
                                      np.array([1.,1.,1.], dtype=float))
        self.agent_orientation = \
            kwargs.pop("model_orientation",np.array([0, 90, 90],
                       dtype=float))
        self.agent_translation = kwargs.pop("model_translation",
                                            np.array([0., 0.], dtype=float))
        self.range_for_zoom = kwargs.pop("is_zoom_range", False)
        self.line_thicknesses = kwargs.pop("line_thickness", { # Dict of keys cameras and values [line thickness, height] which are needed to calculate the dynamic thickness
            -3: [0.03, 10],
            -2: [0.03, 10],
            -1: [0.35, 700],
            0: [0.05, 70]
        })
        # Parameter for agent camera views
        # [driving direction offset, height, x-orientation,y-orientation,z-orientation offset to driving direction, LookAt?]
        self.agent_cam_parameter = kwargs.pop(
            "agent_view", {
                "bird_agent": [0, 70, 0, 270, 270, False],
                "third": [-35, 15, 0, 0, 0, True],
                "first": [1.2, 2.7, 180, 190, 90, False]
            })

        # Set Window Size
        self.wp = WindowProperties().getDefault()
        self.wp.setSize(self.screen_dims[0], self.screen_dims[1])
        self.wp.setTitle("BARK Panda3d Viewer")
        WindowProperties.setDefault(self.wp)
        ShowBase.__init__(self)

        # -2 : global camera with camera control, -1: transition state for enabling mouse control, 0: agent cameras
        self.perspectives = {
            -3: ["third", "autozoom"],
            -2: ["bird_agent", "autozoom"],
            -1: ["None"],
            0: ["bird_agent", "third", "first"]
        }
        

        # Render Objects Dict
        self.agent_nodes = {}
        # Agent Poses Dict
        self.agent_poses = {}
        self.setLight()

        # Creating a plane as floor #TODO parameter from map parameter
        frame = [
            -self.screen_dims[0] / 2, self.screen_dims[0] / 2,
            -self.screen_dims[1] / 2, self.screen_dims[1] / 2
        ]
        self.createPlane(frame=frame, color=VBase4(1, 1, 1, 1))

        # Set up the camera loop task
        self.camera_list = [-1]
        self.initCam()
        self.taskMgr.add(self.spinCameraTask, "SpinCameraTask")

        # Set up the line generator
        self.setDrawer(budget=100000)

        self.perspective = self.perspectives[self.camIndex(self.perspective)]
        self.line_thickness = self.line_thicknesses[self.camIndex(self.perspective)][0]
        self.addButtons()

    def setDrawer(self, budget=100000):
        """Initializes the MeshDrawer() generator to draw lines in Panda3d

        Keyword Arguments:
            budget {int} -- maximum triangles for rendering the mesh (default: {100000})
        """
        self.generator = MeshDrawer()
        self.generator.setBudget(budget)
        self.generatorNode = self.generator.getRoot()
        self.generatorNode.reparentTo(self.render)

    def createPlane(self, frame=None, color=VBase4(1, 1, 11, 1)):
        """ Creates a Plane/Card with the Panda3d Cardmaker() class
        Keyword Arguments:
            frame {list} -- The coordinates [x1,y1,x2,y2] of the planes/cards edges (default: {[-1, -1, 1, 1]})
            color {VBase4} -- The color of the planes/cards (default: {VBase4(1, 1, 1, 1)})
        """
        frame = frame or [-1, -1, 1, 1]
        card = CardMaker("plane")
        card.set_color(color)
        card.set_frame(frame[0], frame[1], frame[2], frame[3])
        n = NodePath()
        self.plane = n.attach_new_node(card.generate())
        self.plane.setHpr(0, 270, 0)
        self.plane.reparentTo(self.render)


    def setLight(self, color=VBase4(.1, .1, .1, 1)):
        """Sets an ambient and omnidirectional light for rendering
        Keyword Arguments:
            color {VBase4} -- color of the ambient light (default: {VBase4(1, 1, 1, 1)})
        """
        alight = AmbientLight('alight')
        alight.setColor(color)
        alnp = self.render.attachNewNode(alight)
        self.render.setLight(alnp)

        # TODO(@hart): position of pt-light needs to be fixed
        plight = PointLight('plight')
        plight.setColor((1, 1, 1, 1))
        self.plnp = self.render.attachNewNode(plight)
        self.plnp.setPos(0, 0, 10000)
        self.render.setLight(self.plnp)


    def addButtons(self):
        self.cam_btn = DirectButton(
            text="Camera",
            scale=0.05,
            command=self.switchCamera,
            pos=Vec3(0.9, 0, 0.9))
        self.per_btn = DirectButton(
            text="View",
            scale=0.05,
            command=self.switchView,
            pos=Vec3(0.9, 0, 0.85))

    def camIndex(self, follow_id):
        return follow_id

    def switchCamera(self):
        # Switches between global and agent cameras
        self.follow_agent_id = self.camera_list.pop(0)
        self.camera_list.append(self.follow_agent_id)
        self.perspective = self.perspectives[self.camIndex(
            self.follow_agent_id)].copy()
        self.line_thickness = self.line_thicknesses[self.camIndex(
            self.follow_agent_id)][0]

    def switchView(self):
        # Switches between the perspectives defined by the camera
        self.perspective.append(self.perspective[0])
        self.perspective.pop(0)

    def updateCamera(self, lookAt=False):
        """Updates the camera position from calls using variables self.cam_pose and self.cam_or

        Keyword Arguments:
            lookAt {bool} -- If true the orientation is calculated by the agents position (default: {False})
        """

        self.camera.setPos(self.cam_pose[0], self.cam_pose[1],
                           self.cam_pose[2])
        if lookAt:
            self.camera.lookAt(self.agent_nodes[self.follow_agent_id])
        else:
            self.camera.setHpr(self.cam_or[2], self.cam_or[1], self.cam_or[0])

    def spinCameraTask(self, task):
        """This function sets and updates the camera according to environment variables and parameters

        Arguments:
            task {task} -- Panda3d Task

        Returns:
            Task.cont -- Panda3d Task Return
        """
        if self.follow_agent_id is -1:  # Transition into camera control view
            self.setMouseControl()
        if self.follow_agent_id is -2:  # Global Camera
            if (self.perspective[0] is self.perspectives[-2][1]):
                self.setAutoZoomCam(self.agent_poses)
        if self.follow_agent_id >= 0:  # Camera for all agents
            self.setAgentCam(self.perspective[0],
                             self.agent_poses[self.follow_agent_id])
        return Task.cont
    
    def initCam(self, poses=None, orientation=None):#TODO Calculate from map parameter before
        poses = poses or [0, 0, 700]
        orientation = orientation or [0, 270, 0]
        self.cam_pose = np.array(poses, dtype=float)
        self.cam_or = np.array(orientation, dtype=float)

    def setMouseControl(self):
        # Enables Panda3d Standard Mouse Control
        # Right mouse button: zoom, left mouse Button: move, middle mouse button: tilt
        self.initCam()
        self.updateCamera(lookAt=False)
        self.mat = Mat4(self.camera.getMat())
        self.mat.invertInPlace()
        base.mouseInterfaceNode.setMat(self.mat)
        base.enableMouse()
        # self.follow_agent_id = -2
        self.perspective = self.perspectives[self.camIndex(
            self.follow_agent_id)].copy()

    def setAutoZoomCam(self, agent_poses):
        """This function calculates the camera position so that all agents are visible

        Arguments:
            agent_poses {[dict]} -- [look up table of all agent poses]
        """
        points = np.array([[]], dtype=float)
        if ((self.world_x_range is not None)
                and (self.world_y_range is not None) and self.range_for_zoom):
            points = np.array([[self.world_x_range[0], self.world_y_range[0]],
                               [self.world_x_range[1], self.world_y_range[1]]])
        # Build concat matrix of all positions
        for _, agent in agent_poses.items():
            if points.size is 0:
                points = np.array([[agent[0], agent[1]]], dtype=float)
            else:
                points = np.append(points, [[agent[0], agent[1]]], axis=0)
        # Calculate maximum distance between min,max of x or y poses
        max_dist = np.max(np.max(points, axis=0) - np.min(points, axis=0))
        # Calculate mean positions as center for camera
        center = (np.max(points, axis=0) + np.min(points, axis=0)) / 2
        # Lower bound for vieweable area
        max_dist = np.maximum(max_dist * 1.25, 40)
        # Calculate camera height
        zoom = max_dist * np.tan(pi * (60 / 180))
        self.cam_pose = np.array([center[0], center[1], zoom], dtype=float)
        self.cam_or = np.array([0, 270, 0], dtype=float)
        self.updateCamera(lookAt=False)

    def setAgentCam(self, perspective, agent_poses):
        """Sets up the class variable for the camerea position and orientation

        Arguments:
            perspective {list} -- List of strings describing the perspective currently ["bird_agent","third","first"] which is used as an index for the parameters dict
            agent_poses {np.array([x,y,theta])} -- Describes the agent position
        """

        base.disableMouse()
        self.cam_pose[0] = agent_poses[0] + self.agent_cam_parameter[
            perspective][0] * np.cos(agent_poses[2])
        self.cam_pose[1] = agent_poses[1] + self.agent_cam_parameter[
            perspective][0] * np.sin(agent_poses[2])
        self.cam_pose[2] = self.agent_cam_parameter[perspective][1]
        self.cam_or = np.array([
            self.agent_cam_parameter[perspective][2],
            self.agent_cam_parameter[perspective][3],
            (agent_poses[2] * 180) / pi +
            self.agent_cam_parameter[perspective][4]
        ],
                               dtype=float)
        self.updateCamera(lookAt=self.agent_cam_parameter[perspective][5])

    def calcLineThickness(self,cameras=None):
        """Uses a linear approximation between two cameras to calculate the line thickness

        Arguments:
            cameras {int array} -- Sets the indices of the target cameras
        Returns:
            [float] -- [The approximated thickness]
        """
        cameras = cameras or [-1,0]
        incline = (self.line_thicknesses[cameras[0]][0] - self.line_thicknesses[cameras[1]][0]) / (self.line_thicknesses[cameras[0]][1]-self.line_thicknesses[cameras[1]][1])
        return np.maximum((self.line_thicknesses[cameras[1]][0] + incline * (self.camera.getZ() - self.line_thicknesses[cameras[1]][1]) ), 0.01)

    def drawPolygon2d(self, polygon, color, alpha, facecolor=None):
        # Draws a polygon with drawLine2d
        points = polygon.ToArray()
        # self.line_thickness = self.calcLineThickness()
        for point in points:
            self.generator.link_segment(
                Vec3(point[0], point[1], 2), Vec4(0, 0, 1, 1),
                self.line_thickness, self.getColor(color))
        self.generator.link_segment_end(Vec4(0, 0, 1, 1), self.getColor(color))


    # def drawLine2d(self, line2d, color=Viewer.Color.Blue, alpha=1.0, zorder=1, dashed=False):
    def drawLine2d(self, line2d, color='blue', alpha=1.0, dashed=False, zorder=1, linewidth=1):
        # TODO: enable dashed line
        line2d_np = line2d.ToArray()
        for point in line2d_np:
            self.generator.link_segment(
                Vec3(point[0], point[1], 2), Vec4(0, 0, 1, 1),
                self.line_thickness, self.getColor(color))
        self.generator.link_segment_end(Vec4(0, 0, 1, 1), self.getColor(color))
    
    def get_aspect_ratio(self):
      # ratio is width / height
      return 1.
    
    def drawAgent(self, agent, color, alpha, facecolor):
        """Draws an agent object with a model previosly set in set with self.agent_model_path
        Arguments:
            agent -- Agent object from world
        """
        # Adds new agent to dict of Panda3d nodes
        if not agent.id in self.agent_nodes:
            self.agent_nodes[agent.id] = self.loader.loadModel(
                self.agent_model_path)
            self.agent_nodes[agent.id].reparentTo(self.render)
            self.agent_nodes[agent.id].setPos(0, 0, 2)
            self.agent_nodes[agent.id].setScale(
                self.agent_scale[0], self.agent_scale[1], self.agent_scale[2])
            self.agent_poses[agent.id] = np.array([0, 0, 0], dtype=float)

            self.camera_list.insert(len(self.camera_list), agent.id)

        # Set new pose of agent in world cordinates
        state = agent.followed_trajectory[-1, :]
        self.agent_poses[agent.id][0] = state[int(StateDefinition.X_POSITION)]
        self.agent_poses[agent.id][1] = state[int(StateDefinition.Y_POSITION)]
        self.agent_poses[agent.id][2] = state[int(
            StateDefinition.THETA_POSITION)]
        transformed_polygon = agent.shape.Transform(self.agent_poses[agent.id])
        # self.drawPolygon2d(transformed_polygon, self.getColor(Viewer.Color.Blue), 1.0)

        # Fitting of 3d-Model frame to agent positon
        angle = ((self.agent_poses[agent.id][2] * 180) / pi +
                 self.agent_orientation[0])
                 
        self.agent_nodes[agent.id].setHpr(angle, self.agent_orientation[1],
                                          self.agent_orientation[2])
        
        translation = self.agent_translation
        if not np.all(translation == 0):
            r = np.array([[
                np.cos(self.agent_poses[agent.id][2]),
                np.sin(self.agent_poses[agent.id][2])
            ],
            [
                np.cos(self.agent_poses[agent.id][2] + pi * 0.5),
                np.sin(self.agent_poses[agent.id][2] + pi * 0.5)
            ]],
                         dtype=float)
            translation = np.dot(translation, r)
        self.agent_nodes[agent.id].setPos(
            self.agent_poses[agent.id][0] + translation[0],
            self.agent_poses[agent.id][1] + translation[1], 2)


    def getColor(self, color):
        if isinstance(color, Viewer.Color):
            return {
                Viewer.Color.White:
                Vec4(1, 1, 1, 1),
                Viewer.Color.Red:
                Vec4(1, 0, 0, 1),
                Viewer.Color.Blue:
                Vec4(0, 0, 1, 1),
                Viewer.Color.Magenta:
                Vec4(100.0 / 255.0, 1, 100.0 / 255.0, 1),
                Viewer.Color.Brown:
                Vec4(150.0 / 255.0, 50.0 / 255.0, 50.0 / 255.0, 1),
                Viewer.Color.Black:
                Vec4(0, 0, 0, 1)
            }.get(color, Vec4(0, 0, 0, 1))
        else:
            return Vec4(color[0], color[1], color[2], 1)

    def drawWorld(self, world, eval_agent_ids=None, filename=None, scenario_idx=None, debug_text=True):
        self.generator.begin(base.cam, self.render)
        super(Panda3dViewer, self).drawWorld(world, eval_agent_ids, filename, scenario_idx, debug_text)
        self.generator.end()
        self.taskMgr.step()

    def Reset(self):
        # for key, agent in self.agent_nodes.items():
        #   agent.removeNode()
        # self.agent_nodes = {}
        # self.follow_agent_id = 
        for an in self.agent_nodes.values():
          an.removeNode()
        self.agent_poses = {}



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.


import numpy as np
import time
import unittest
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.commons.xodr_parser import XodrParser
from bark.core.models.behavior import BehaviorConstantVelocity, \
  BehaviorMPContinuousActions
from bark.core.models.execution import ExecutionModelInterpolate
from bark.core.models.dynamic import SingleTrackModel, StateDefinition
from bark.core.world import World
from bark.core.world.goal_definition import GoalDefinitionPolygon, \
  GoalDefinitionStateLimits, GoalDefinitionSequential, \
    GoalDefinitionStateLimitsFrenet
from bark.core.world.agent import Agent
from bark.core.world.map import MapInterface
from bark.core.geometry.standard_shapes import CarLimousine
from bark.core.geometry import Point2d, Polygon2d, Line2d
from bark.core.world.evaluation import EvaluatorGoalReached, \
  EvaluatorCollisionEgoAgent, EvaluatorStepCount



class EvaluationTests(unittest.TestCase):
  def test_one_agent_at_goal_polygon(self):
    param_server = ParameterServer()
    # Model Definition
    behavior_model = BehaviorConstantVelocity(param_server)
    execution_model = ExecutionModelInterpolate(param_server)
    dynamic_model = SingleTrackModel(param_server)

    # Agent Definition
    agent_2d_shape = CarLimousine()
    init_state = np.array([0, -191.789,-50.1725, 3.14*3.0/4.0, 150/3.6])
    agent_params = param_server.AddChild("agent1")
    goal_polygon = Polygon2d([0, 0, 0],
                             [Point2d(-4,-4),
                              Point2d(-4,4),
                              Point2d(4,4),
                              Point2d(4,-4)])
    goal_polygon = goal_polygon.Translate(Point2d(-191.789,-50.1725))

    agent = Agent(init_state,
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                GoalDefinitionPolygon(goal_polygon),
                  None)

    world = World(param_server)
    world.AddAgent(agent)
    evaluator = EvaluatorGoalReached(agent.id)
    world.AddEvaluator("success", evaluator)


    info = world.Evaluate()
    self.assertEqual(info["success"], True)

  def test_one_agent_at_goal_state_limits(self):
    param_server = ParameterServer()
    # Model Definition
    behavior_model = BehaviorConstantVelocity(param_server)
    execution_model = ExecutionModelInterpolate(param_server)
    dynamic_model = SingleTrackModel(param_server)

    # Agent Definition
    agent_2d_shape = CarLimousine()
    init_state = np.array([0, -191.789,-50.1725, 3.14*3.0/4.0, 150/3.6])
    agent_params = param_server.AddChild("agent1")
    goal_polygon = Polygon2d([0, 0, 0],
                             [Point2d(-1,-1),
                              Point2d(-1,1),
                              Point2d(1,1),
                              Point2d(1,-1)])
    goal_polygon = goal_polygon.Translate(Point2d(-191.789,-50.1725))

    agent = Agent(init_state,
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                GoalDefinitionStateLimits(goal_polygon, (3.14*3.0/4.0-0.08, 3.14*3.0/4.0+0.08)),
                  None)

    world = World(param_server)
    world.AddAgent(agent)
    evaluator = EvaluatorGoalReached(agent.id)
    world.AddEvaluator("success", evaluator)


    info = world.Evaluate()
    self.assertEqual(info["success"], True)

  def test_one_agent_at_goal_sequential(self):
    param_server = ParameterServer()
    # Model Definition
    dynamic_model = SingleTrackModel(param_server)
    behavior_model = BehaviorMPContinuousActions(param_server)
    idx = behavior_model.AddMotionPrimitive(np.array([1, 0]))
    behavior_model.ActionToBehavior(idx)
    execution_model = ExecutionModelInterpolate(param_server)


    # Agent Definition
    agent_2d_shape = CarLimousine()
    init_state = np.array([0, 0, 0, 0, 0])
    agent_params = param_server.AddChild("agent1")
    goal_frame = Polygon2d([0, 0, 0],
                             [Point2d(-1,-1),
                              Point2d(-1,1),
                              Point2d(1,1),
                              Point2d(1,-1)])

    goal_polygon1 = goal_frame.Translate(Point2d(10, 0))
    goal_polygon2 = goal_frame.Translate(Point2d(20, 0))
    goal_polygon3 = goal_frame.Translate(Point2d(30, 0))

    goal_def1 = GoalDefinitionStateLimits(goal_polygon1, [-0.08, 0.08])
    goal_def2 = GoalDefinitionStateLimits(goal_polygon2, [-0.08, 0.08])
    goal_def3 = GoalDefinitionStateLimits(goal_polygon3, [-0.08, 0.08])

    goal_definition = GoalDefinitionSequential([goal_def1,
                                                goal_def2,
                                                goal_def3])

    self.assertEqual(len(goal_definition.sequential_goals),3)
    agent = Agent(init_state,
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)

    world = World(param_server)
    world.AddAgent(agent)
    evaluator = EvaluatorGoalReached(agent.id)
    world.AddEvaluator("success", evaluator)

    # just drive with the single motion primitive should be successful 
    for _ in range(0,1000):
        world.Step(0.2)
        info = world.Evaluate()
        if info["success"]:
            break
    
    self.assertEqual(info["success"], True)
    self.assertAlmostEqual(agent.state[int(StateDefinition.X_POSITION)], 30, delta=0.5)
      

  def test_one_agent_at_goal_state_limits_frenet(self):
    param_server = ParameterServer()
    # Model Definition
    behavior_model = BehaviorConstantVelocity(param_server)
    execution_model = ExecutionModelInterpolate(param_server)
    dynamic_model = SingleTrackModel(param_server)

    # Agent Definition
    agent_2d_shape = CarLimousine()
    agent_params = param_server.AddChild("agent1")

    center_line = Line2d()
    center_line.AddPoint(Point2d(5.0, 5.0))
    center_line.AddPoint(Point2d(10.0, 10.0))
    center_line.AddPoint(Point2d(20.0, 10.0))

    max_lateral_dist = (0.4,1)
    max_orientation_diff = (0.08, 0.1)
    velocity_range = (20.0, 25.0)
    goal_definition = GoalDefinitionStateLimitsFrenet(center_line,
                    max_lateral_dist, max_orientation_diff,
                    velocity_range)

    # not at goal x,y, others yes
    agent1 = Agent(np.array([0, 6, 8, 3.14/4.0 , velocity_range[0]]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)

    # at goal x,y and others
    agent2 = Agent(np.array([0, 5.0, 5.5, 3.14/4.0 , velocity_range[1]]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)

    # not at goal x,y,v yes but not orientation
    agent3 = Agent(np.array([0, 5, 5.5, 3.14/4.0+max_orientation_diff[1]+0.001 , 20]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)


    # not at goal x,y, orientation but not v
    agent4 = Agent(np.array([0, 5, 4.5, 3.14/4-max_orientation_diff[0], velocity_range[0]-0.01]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)

    # at goal x,y, at lateral limit 
    agent5 = Agent(np.array([0, 15, 10-max_lateral_dist[0]+0.05, 0, velocity_range[1]]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)

    # not at goal x,y slightly out of lateral limit 
    agent6 = Agent(np.array([0, 15, 10+max_lateral_dist[0]+0.05, 3.14/4+max_orientation_diff[0], velocity_range[0]]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)


    # not at goal x,y,v yes but not orientation
    agent7 = Agent(np.array([0, 5, 5.5, 3.14/4.0-max_orientation_diff[0]-0.001 , 20]),
                behavior_model,
                dynamic_model,
                execution_model,
                agent_2d_shape,
                agent_params,
                goal_definition,
                  None)


    world = World(param_server)
    world.AddAgent(agent1)
    world.AddAgent(agent2)
    world.AddAgent(agent3)
    world.AddAgent(agent4)
    world.AddAgent(agent5)
    world.AddAgent(agent6)
    world.AddAgent(agent7)

    evaluator1 = EvaluatorGoalReached(agent1.id)
    evaluator2 = EvaluatorGoalReached(agent2.id)
    evaluator3 = EvaluatorGoalReached(agent3.id)
    evaluator4 = EvaluatorGoalReached(agent4.id)
    evaluator5 = EvaluatorGoalReached(agent5.id)
    evaluator6 = EvaluatorGoalReached(agent6.id)
    evaluator7 = EvaluatorGoalReached(agent7.id)
    world.AddEvaluator("success1", evaluator1)
    world.AddEvaluator("success2", evaluator2)
    world.AddEvaluator("success3", evaluator3)
    world.AddEvaluator("success4", evaluator4)
    world.AddEvaluator("success5", evaluator5)
    world.AddEvaluator("success6", evaluator6)
    world.AddEvaluator("success7", evaluator7)


    info = world.Evaluate()
    self.assertEqual(info["success1"], False)
    self.assertEqual(info["success2"], True)
    self.assertEqual(info["success3"], False)
    self.assertEqual(info["success4"], False)
    self.assertEqual(info["success5"], True)
    self.assertEqual(info["success6"], False)
    self.assertEqual(info["success7"], False)
        
if __name__ == '__main__':
  unittest.main()




from .matplotlib_viewer import MPViewer
from .pygame_viewer import PygameViewer
from .panda3d_viewer import Panda3dViewer
from .video_renderer import VideoRenderer
from .viewer import BaseViewer

__all__ = ["MPViewer","PygameViewer","Panda3dViewer","VideoRenderer","BaseViewer"]



# Copyright (c) 2020 Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.


import time

def return_execution_time(func):
  def wrapper(*args, **kwargs):
    time_before_step = time.time()
    func(*args, **kwargs)
    step_duration = time.time() - time_before_step
    return step_duration

  return wrapper